/*
 *  linux/arch/i386/mm/pgtable.c
 */

#include <linux/config.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/mm.h>
#include <linux/swap.h>
#include <linux/smp.h>
#include <linux/highmem.h>
#include <linux/slab.h>
#include <linux/pagemap.h>

#include <asm/system.h>
#include <asm/pgtable.h>
#include <asm/pgalloc.h>
#include <asm/fixmap.h>
#include <asm/e820.h>
#include <asm/tlb.h>
#include <asm/tlbflush.h>
#include <asm/io.h>

#include <asm-xen/foreign_page.h>

void show_mem(void)
{
	int total = 0, reserved = 0;
	int shared = 0, cached = 0;
	int highmem = 0;
	struct page *page;
	pg_data_t *pgdat;
	unsigned long i;

	printk("Mem-info:\n");
	show_free_areas();
	printk("Free swap:       %6dkB\n",nr_swap_pages<<(PAGE_SHIFT-10));
	for_each_pgdat(pgdat) {
		for (i = 0; i < pgdat->node_spanned_pages; ++i) {
			page = pgdat->node_mem_map + i;
			total++;
			if (PageHighMem(page))
				highmem++;
			if (PageReserved(page))
				reserved++;
			else if (PageSwapCache(page))
				cached++;
			else if (page_count(page))
				shared += page_count(page) - 1;
		}
	}
	printk("%d pages of RAM\n", total);
	printk("%d pages of HIGHMEM\n",highmem);
	printk("%d reserved pages\n",reserved);
	printk("%d pages shared\n",shared);
	printk("%d pages swap cached\n",cached);
}

/*
 * Associate a virtual page frame with a given physical page frame 
 * and protection flags for that frame.
 */ 
static void set_pte_pfn(unsigned long vaddr, unsigned long pfn, pgprot_t flags)
{
	pgd_t *pgd;
	pmd_t *pmd;
	pte_t *pte;

	pgd = swapper_pg_dir + pgd_index(vaddr);
	if (pgd_none(*pgd)) {
		BUG();
		return;
	}
	pmd = pmd_offset(pgd, vaddr);
	if (pmd_none(*pmd)) {
		BUG();
		return;
	}
	pte = pte_offset_kernel(pmd, vaddr);
	/* <pfn,flags> stored as-is, to permit clearing entries */
	set_pte(pte, pfn_pte(pfn, flags));

	/*
	 * It's enough to flush this one mapping.
	 * (PGE mappings get flushed as well)
	 */
	__flush_tlb_one(vaddr);
}

/*
 * Associate a virtual page frame with a given physical page frame 
 * and protection flags for that frame.
 */ 
static void set_pte_pfn_ma(unsigned long vaddr, unsigned long pfn,
			   pgprot_t flags)
{
	pgd_t *pgd;
	pmd_t *pmd;
	pte_t *pte;

	pgd = swapper_pg_dir + pgd_index(vaddr);
	if (pgd_none(*pgd)) {
		BUG();
		return;
	}
	pmd = pmd_offset(pgd, vaddr);
	if (pmd_none(*pmd)) {
		BUG();
		return;
	}
	pte = pte_offset_kernel(pmd, vaddr);
	/* <pfn,flags> stored as-is, to permit clearing entries */
	set_pte(pte, pfn_pte_ma(pfn, flags));

	/*
	 * It's enough to flush this one mapping.
	 * (PGE mappings get flushed as well)
	 */
	__flush_tlb_one(vaddr);
}

/*
 * Associate a large virtual page frame with a given physical page frame 
 * and protection flags for that frame. pfn is for the base of the page,
 * vaddr is what the page gets mapped to - both must be properly aligned. 
 * The pmd must already be instantiated. Assumes PAE mode.
 */ 
void set_pmd_pfn(unsigned long vaddr, unsigned long pfn, pgprot_t flags)
{
	pgd_t *pgd;
	pmd_t *pmd;

	if (vaddr & (PMD_SIZE-1)) {		/* vaddr is misaligned */
		printk ("set_pmd_pfn: vaddr misaligned\n");
		return; /* BUG(); */
	}
	if (pfn & (PTRS_PER_PTE-1)) {		/* pfn is misaligned */
		printk ("set_pmd_pfn: pfn misaligned\n");
		return; /* BUG(); */
	}
	pgd = swapper_pg_dir + pgd_index(vaddr);
	if (pgd_none(*pgd)) {
		printk ("set_pmd_pfn: pgd_none\n");
		return; /* BUG(); */
	}
	pmd = pmd_offset(pgd, vaddr);
	set_pmd(pmd, pfn_pmd(pfn, flags));
	/*
	 * It's enough to flush this one mapping.
	 * (PGE mappings get flushed as well)
	 */
	__flush_tlb_one(vaddr);
}

void __set_fixmap (enum fixed_addresses idx, unsigned long phys, pgprot_t flags)
{
	unsigned long address = __fix_to_virt(idx);

	if (idx >= __end_of_fixed_addresses) {
		BUG();
		return;
	}
	set_pte_pfn(address, phys >> PAGE_SHIFT, flags);
}

void __set_fixmap_ma (enum fixed_addresses idx, unsigned long phys, pgprot_t flags)
{
	unsigned long address = __fix_to_virt(idx);

	if (idx >= __end_of_fixed_addresses) {
		BUG();
		return;
	}
	set_pte_pfn_ma(address, phys >> PAGE_SHIFT, flags);
}

pte_t *pte_alloc_one_kernel(struct mm_struct *mm, unsigned long address)
{
	pte_t *pte = (pte_t *)__get_free_page(GFP_KERNEL|__GFP_REPEAT);
	if (pte) {
		clear_page(pte);
		make_page_readonly(pte);
		xen_flush_page_update_queue();
	}
	return pte;
}

void pte_ctor(void *pte, kmem_cache_t *cache, unsigned long unused)
{
	struct page *page = virt_to_page(pte);
	SetPageForeign(page, pte_free);
	set_page_count(page, 1);

	clear_page(pte);
	make_page_readonly(pte);
	queue_pte_pin(__pa(pte));
	flush_page_update_queue();
}

void pte_dtor(void *pte, kmem_cache_t *cache, unsigned long unused)
{
	struct page *page = virt_to_page(pte);
	ClearPageForeign(page);

	queue_pte_unpin(__pa(pte));
	make_page_writable(pte);
	flush_page_update_queue();
}

struct page *pte_alloc_one(struct mm_struct *mm, unsigned long address)
{
	pte_t *ptep;

#ifdef CONFIG_HIGHPTE
	struct page *pte;

	pte = alloc_pages(GFP_KERNEL|__GFP_HIGHMEM|__GFP_REPEAT, 0);
	if (pte == NULL)
		return pte;
	if (pte >= highmem_start_page) {
		clear_highpage(pte);
		return pte;
	}
	/* not a highmem page -- free page and grab one from the cache */
	__free_page(pte);
#endif
	ptep = kmem_cache_alloc(pte_cache, GFP_KERNEL);
	if (ptep)
		return virt_to_page(ptep);
	return NULL;
}

void pte_free(struct page *pte)
{
	set_page_count(pte, 1);
#ifdef CONFIG_HIGHPTE
	if (pte < highmem_start_page)
#endif
		kmem_cache_free(pte_cache,
				phys_to_virt(page_to_pseudophys(pte)));
#ifdef CONFIG_HIGHPTE
	else
		__free_page(pte);
#endif
}

void pgd_ctor(void *pgd, kmem_cache_t *cache, unsigned long unused)
{

	memcpy((pgd_t *)pgd + USER_PTRS_PER_PGD,
			swapper_pg_dir + USER_PTRS_PER_PGD,
			(PTRS_PER_PGD - USER_PTRS_PER_PGD) * sizeof(pgd_t));

	memset(pgd, 0, USER_PTRS_PER_PGD*sizeof(pgd_t));

	make_page_readonly(pgd);
	queue_pgd_pin(__pa(pgd));
	flush_page_update_queue();
}

void pgd_dtor(void *pgd, kmem_cache_t *cache, unsigned long unused)
{

	queue_pgd_unpin(__pa(pgd));
	make_page_writable(pgd);
	flush_page_update_queue();
}

#ifdef CONFIG_X86_PAE

pgd_t *pgd_alloc(struct mm_struct *mm)
{
	int i;
	pgd_t *pgd = kmem_cache_alloc(pae_pgd_cachep, GFP_KERNEL);

	if (pgd) {
		for (i = 0; i < USER_PTRS_PER_PGD; i++) {
			unsigned long pmd = __get_free_page(GFP_KERNEL);
			if (!pmd)
				goto out_oom;
			clear_page(pmd);
			set_pgd(pgd + i, __pgd(1 + __pa(pmd)));
		}
		memcpy(pgd + USER_PTRS_PER_PGD,
			swapper_pg_dir + USER_PTRS_PER_PGD,
			(PTRS_PER_PGD - USER_PTRS_PER_PGD) * sizeof(pgd_t));
	}
	return pgd;
out_oom:
	for (i--; i >= 0; i--)
		free_page((unsigned long)__va(pgd_val(pgd[i])-1));
	kmem_cache_free(pae_pgd_cachep, pgd);
	return NULL;
}

void pgd_free(pgd_t *pgd)
{
	int i;

	for (i = 0; i < USER_PTRS_PER_PGD; i++)
		free_page((unsigned long)__va(pgd_val(pgd[i])-1));
	kmem_cache_free(pae_pgd_cachep, pgd);
}

#else

pgd_t *pgd_alloc(struct mm_struct *mm)
{
	pgd_t *pgd = kmem_cache_alloc(pgd_cache, GFP_KERNEL);

	return pgd;
}

void pgd_free(pgd_t *pgd)
{
	kmem_cache_free(pgd_cache, pgd);
}

#endif /* CONFIG_X86_PAE */

void make_lowmem_page_readonly(void *va)
{
	pte_t *pte = virt_to_ptep(va);
	queue_l1_entry_update(pte, (*(unsigned long *)pte)&~_PAGE_RW);
}

void make_lowmem_page_writable(void *va)
{
	pte_t *pte = virt_to_ptep(va);
	queue_l1_entry_update(pte, (*(unsigned long *)pte)|_PAGE_RW);
}

void make_page_readonly(void *va)
{
	pte_t *pte = virt_to_ptep(va);
	queue_l1_entry_update(pte, (*(unsigned long *)pte)&~_PAGE_RW);
	if ( (unsigned long)va >= (unsigned long)high_memory )
	{
		unsigned long phys;
		phys = machine_to_phys(*(unsigned long *)pte & PAGE_MASK);
#ifdef CONFIG_HIGHMEM
		if ( (phys >> PAGE_SHIFT) < highstart_pfn )
#endif
			make_lowmem_page_readonly(phys_to_virt(phys));
	}
}

void make_page_writable(void *va)
{
	pte_t *pte = virt_to_ptep(va);
	queue_l1_entry_update(pte, (*(unsigned long *)pte)|_PAGE_RW);
	if ( (unsigned long)va >= (unsigned long)high_memory )
	{
		unsigned long phys;
		phys = machine_to_phys(*(unsigned long *)pte & PAGE_MASK);
#ifdef CONFIG_HIGHMEM
		if ( (phys >> PAGE_SHIFT) < highstart_pfn )
#endif
			make_lowmem_page_writable(phys_to_virt(phys));
	}
}

void make_pages_readonly(void *va, unsigned int nr)
{
	while ( nr-- != 0 )
	{
		make_page_readonly(va);
		va = (void *)((unsigned long)va + PAGE_SIZE);
	}
}

void make_pages_writable(void *va, unsigned int nr)
{
	while ( nr-- != 0 )
	{
		make_page_writable(va);
		va = (void *)((unsigned long)va + PAGE_SIZE);
	}
}

