/* linux/drivers/block/ckrm_io.c : Block I/O Resource Controller for CKRM
 *
 * Copyright (C) Shailabh Nagar, IBM Corp. 2004
 * 
 * 
 * Provides best-effort block I/O bandwidth control for CKRM 
 * This file provides the CKRM API. The underlying scheduler is the
 * ps (proportional share) ioscheduler.
 *
 * Latest version, more details at http://ckrm.sf.net
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/list.h>
#include <linux/spinlock.h>
#include <linux/fs.h>
#include <linux/parser.h>
#include <linux/kobject.h>
#include <asm/errno.h>
#include <asm/div64.h>

#include <linux/ckrm_tc.h>
#include <linux/ckrm-io.h>

/* sectorate == 512 byte sectors served in PS_EPOCH ns*/

#define CKI_ROOTSECTORATE_DEF	100000
#define CKI_MINSECTORATE_DEF	100

typedef struct ckrm_io_class {

	struct ckrm_core_class *core;
	struct ckrm_core_class *parent;
	
	

	struct ckrm_shares shares;
	spinlock_t  lock; /* protect rate_list and cnt_*  */
	
	struct list_head  rate_list;

	/* Absolute shares of this class
	 * in local units. 
	 */
	int cnt_guarantee; /* Allocation as parent */
	int cnt_unused;    /* Allocation to default subclass */
	int cnt_limit;

} cki_icls_t;

/* Internal functions */
static inline void init_icls_one(cki_icls_t *icls);
static void cki_recalc_propagate(cki_icls_t *res, cki_icls_t *parres);
static int cki_recalc_abs(void);

/* Functions from ps_iosched */
extern int ps_drop_psq(struct ps_data *psd, void *key);


/* CKRM Resource Controller API functions */
static void * cki_alloc(struct ckrm_core_class *this,
			struct ckrm_core_class * parent);
static void cki_free(void *res);
static int cki_setshare(void *res, struct ckrm_shares * shares);
static int cki_getshare(void *res, struct ckrm_shares * shares);
static int cki_getstats(void *res, struct seq_file *);
static int cki_resetstats(void *res);
static int cki_showconfig(void *res, struct seq_file *sfile);
static int cki_setconfig(void *res, const char *cfgstr);
static void cki_chgcls(void *tsk, void *oldres, void *newres);

/* Global data */
struct ckrm_res_ctlr cki_rcbs;

struct cki_data ckid;
EXPORT_SYMBOL_GPL(ckid);

struct ps_rate cki_def_psrate; 
EXPORT_SYMBOL_GPL(cki_def_psrate);

struct rw_semaphore psdlistsem;
EXPORT_SYMBOL(psdlistsem);

LIST_HEAD(ps_psdlist);
EXPORT_SYMBOL(ps_psdlist);


/* Print psdrate entries in icls for all current psd's */
void print_psdlist(void)
{
	struct psd_list_entry *psdl;

	printk(KERN_WARNING "-----> psdlist start\n");
	down_read(&psdlistsem);
	list_for_each_entry(psdl,&ps_psdlist,psd_list) { 
		printk(KERN_WARNING "psdl %p psd %p\n",
		       psdl,psdl->psd);
		
	}
	up_read(&psdlistsem);
	printk(KERN_WARNING "-----> psdlist stop\n");
}

static void cki_setpsdnames(void)
{
        static int invoked=0;
        struct psd_list_entry *psdl;
        char *path;

        if (++invoked > 1)
                return;

        down_read(&psdlistsem);
        list_for_each_entry(psdl,&ps_psdlist,psd_list) {
		if (psdl->psd->queue->kobj.k_name) {
			path = kobject_get_path(&(psdl->psd->queue->kobj), GFP_KERNEL);
			if (path)
				psdl->psd->name = path;
		}
        }
        up_read(&psdlistsem);
}

static int  cki_init(void)
{
	struct ckrm_classtype *clstype;
	int resid = cki_rcbs.resid;

	if (resid != -1) 
		return 0;

	clstype = ckrm_find_classtype_by_name("taskclass");
	if (clstype == NULL) {
		printk(KERN_WARNING "%s: classtype<taskclass> not found\n",
		       __FUNCTION__);
		return -ENOENT;
	}

	ckid.cfglock = SPIN_LOCK_UNLOCKED;
	ckid.rootsectorate = CKI_ROOTSECTORATE_DEF;
	ckid.minsectorate = CKI_MINSECTORATE_DEF;

	atomic_set(&cki_def_psrate.sectorate,0);
	init_rwsem(&psdlistsem);
	
	resid = ckrm_register_res_ctlr(clstype, &cki_rcbs);
	if (resid == -1) 
		return -ENOENT;

	cki_rcbs.classtype = clstype;

	return 0;
}

static void  cki_exit(void)
{
	ckrm_unregister_res_ctlr(&cki_rcbs);
	cki_rcbs.resid = -1;
	cki_rcbs.classtype = NULL; 
}

static struct psdrate *cki_find_rate(struct ckrm_io_class *icls,
				     struct ps_data *psd)
{
	struct psdrate *prate;
	
	spin_lock(&icls->lock);
	list_for_each_entry(prate, &icls->rate_list, rate_list) {
		if (prate->psd == psd)
			goto found;
	}
	prate = NULL;
found:
	spin_unlock(&icls->lock);
	return prate;
}

/* Exported functions */

void __cki_set_sectorate(cki_icls_t *icls, int sectorate)
{
	struct psdrate *prate;
	u64 temp;

	list_for_each_entry(prate, &icls->rate_list, rate_list) {
		temp = (u64) sectorate * prate->psd->ps_max_sectorate;
		do_div(temp,ckid.rootsectorate);
		atomic_set(&prate->psrate.sectorate,temp);
	}	
}	

void cki_set_sectorate(cki_icls_t *icls, int sectorate)
{
	spin_lock(&icls->lock);
	__cki_set_sectorate(icls,sectorate);
	spin_unlock(&icls->lock);
}

/* Reset psdrate entries in icls for all current psd's 
 * Called after a class's absolute shares change 
 */
void cki_reset_sectorate(cki_icls_t *icls)
{
	struct psdrate *prate;
	u64 temp;
	
	spin_lock(&icls->lock);
	list_for_each_entry(prate, &icls->rate_list, rate_list) {

		if (icls->cnt_limit != CKRM_SHARE_DONTCARE) {
			temp = (u64) icls->cnt_limit * prate->psd->ps_max_sectorate;
			do_div(temp,ckid.rootsectorate);
		} else 
			temp = prate->psd->ps_min_sectorate;
		atomic_set(&prate->psrate.sectorate,temp);
	}	
	spin_unlock(&icls->lock);

}

struct psdrate *dbprate;

int cki_psdrate_init(struct ckrm_io_class *icls, struct ps_data *psd)
{
	struct psdrate *prate;
	u64 temp;

	prate = kmalloc(sizeof(struct psdrate),GFP_KERNEL);
	if (!prate) 
		return -ENOMEM;
	
	INIT_LIST_HEAD(&prate->rate_list);
	prate->psd = psd;
	memset(&prate->psrate,0,sizeof(prate->psrate));
	
	dbprate = prate;
	if (icls->cnt_limit != CKRM_SHARE_DONTCARE) {
		temp = (u64) icls->cnt_limit * psd->ps_max_sectorate;
		do_div(temp,ckid.rootsectorate);
	} else { 
		temp = psd->ps_min_sectorate;
	}
	atomic_set(&prate->psrate.sectorate,temp);
	
	spin_lock(&icls->lock);
	list_add(&prate->rate_list,&icls->rate_list);
	spin_unlock(&icls->lock);

	return 0;
}

int cki_psdrate_del(struct ckrm_io_class *icls, struct ps_data *psd)
{
	struct psdrate *prate;

	prate = cki_find_rate(icls, psd);
	if (!prate) 
		return 0;

	spin_lock(&icls->lock);
	list_del(&prate->rate_list);
	spin_unlock(&icls->lock);

	kfree(prate);
	return 0;
}


/* Create psdrate entries in icls for all current psd's */
void cki_rates_init(cki_icls_t *icls)
{
	struct psd_list_entry *psdl;
	
	down_read(&psdlistsem);
	list_for_each_entry(psdl,&ps_psdlist,psd_list) { 
		if (cki_psdrate_init(icls, psdl->psd)) {
			printk(KERN_WARNING "%s: psdrate addition failed\n",
			       __FUNCTION__);
			continue;
		}
	}
	up_read(&psdlistsem);
}

/* Free all psdrate entries in icls */
void cki_rates_del(cki_icls_t *icls)
{
	struct psdrate *prate, *tmp;
	
	spin_lock(&icls->lock);
	list_for_each_entry_safe(prate, tmp, &icls->rate_list, rate_list) {
	    list_del(&prate->rate_list);
	    kfree(prate);
	}
	spin_unlock(&icls->lock);
/*	
	down_read(&psdlistsem);
	list_for_each_entry(psdl,&ps_psdlist,psd_list) { 
		cki_psdrate_del(icls,psdl->psd);
	}
	up_read(&psdlistsem);
*/
}

/* Called from ps-iosched.c when it initializes a new ps_data
 *  as part of starting to manage a new device request queue 
 */

int cki_psd_init(struct ps_data *psd)
{
	struct ckrm_classtype *ctype = ckrm_classtypes[CKRM_CLASSTYPE_TASK_CLASS];
	struct ckrm_core_class *core;
	struct ckrm_io_class *icls;
	struct psdrate *prate;
	int ret=-ENOMEM;

	/* Set psd's min and max sectorate from default values */
	psd->ps_max_sectorate = ckid.rootsectorate;
	psd->ps_min_sectorate = ckid.minsectorate;

	read_lock(&ckrm_class_lock);
	list_for_each_entry(core, &ctype->classes, clslist) {
		icls = ckrm_get_res_class(core, cki_rcbs.resid, cki_icls_t);
		if (!icls) {
			printk(KERN_WARNING "%s: no icls found for core %p <%s>\n",
			       __FUNCTION__,core,core->name);
			continue;
		}

		prate = cki_find_rate(icls, psd);
		if (prate) 
			continue;

		if (cki_psdrate_init(icls, psd)) {
			printk(KERN_WARNING "%s: psdrate addition failed\n",
			       __FUNCTION__);
			continue;
		}
	}
	ret = 0;

	read_unlock(&ckrm_class_lock);
	return ret;
}
EXPORT_SYMBOL_GPL(cki_psd_init);

/* Called whenever ps-iosched frees a ps_data 
 *  as part of ending management of a device request queue 
 */

int cki_psd_del(struct ps_data *psd)
{
	struct ckrm_classtype *ctype = ckrm_classtypes[CKRM_CLASSTYPE_TASK_CLASS];
	struct ckrm_core_class *core;
	struct ckrm_io_class *icls;
	int ret = 0;

	read_lock(&ckrm_class_lock);
	list_for_each_entry(core, &ctype->classes, clslist) {
		icls = ckrm_get_res_class(core, cki_rcbs.resid, cki_icls_t);
		if (!icls)
			continue;

		if (cki_psdrate_del(icls,psd)) {
			printk(KERN_WARNING "%s: psdrate deletion failed\n",
			       __FUNCTION__);
			continue;
		}
	}
	read_unlock(&ckrm_class_lock);
	return ret;
}
EXPORT_SYMBOL_GPL(cki_psd_del);

void cki_psdlist_add(struct psd_list_entry *psdl) 
{
	down_write(&psdlistsem);
	list_add(&psdl->psd_list,&ps_psdlist);
	up_write(&psdlistsem);
}

void cki_psdlist_del(struct ps_data *psd) 
{
	struct psd_list_entry *psdl,*tmp;

	down_write(&psdlistsem);
	list_for_each_entry_safe(psdl,tmp,&ps_psdlist,psd_list) {
		if (psdl->psd == psd) {
			list_del(&psdl->psd_list);
			kfree(psdl);
			break;
		}
	}
	up_write(&psdlistsem);
}


struct ps_rate *cki_tsk_psrate(struct ps_data *psd, struct task_struct *tsk)
{
	cki_icls_t *icls;
	struct psdrate *prate;

	icls = ckrm_get_res_class(class_core(tsk->taskclass),
				  cki_rcbs.resid, cki_icls_t);
	if (!icls)
		return NULL;
	
	
	prate = cki_find_rate(icls,psd);
	if (prate)
	    return &(prate->psrate);
	else
	    return NULL;
}
EXPORT_SYMBOL_GPL(cki_tsk_psrate);			

/* Exported functions end */

/* Initialize icls to default values 
 * No other classes touched, locks not reinitialized.
 */

static inline void init_icls_one(cki_icls_t *icls)
{
	/* Zero initial guarantee for scalable creation of
	   multiple classes */

	/* Try out a new set */
	
	icls->shares.my_guarantee = CKRM_SHARE_DONTCARE;
	icls->shares.my_limit = CKRM_SHARE_DONTCARE;
	icls->shares.total_guarantee = CKRM_SHARE_DFLT_TOTAL_GUARANTEE;
	icls->shares.max_limit = CKRM_SHARE_DFLT_MAX_LIMIT;
	icls->shares.unused_guarantee = icls->shares.total_guarantee;
	icls->shares.cur_max_limit = 0;

	icls->cnt_guarantee = CKRM_SHARE_DONTCARE;
	icls->cnt_unused = CKRM_SHARE_DONTCARE;
	icls->cnt_limit = CKRM_SHARE_DONTCARE;

	INIT_LIST_HEAD(&icls->rate_list);
}

/* Initialize root's psd entries */
static void cki_createrootrate(cki_icls_t *root, int sectorate)
{
	spin_lock(&root->lock);
	root->cnt_guarantee = sectorate;
	root->cnt_unused = sectorate;
	root->cnt_limit = sectorate;
	spin_unlock(&root->lock);

	cki_rates_init(root);
}

/* Called with root->share_lock held  */
static void cki_setrootrate(cki_icls_t *root, int sectorate)
{
	spin_lock(&root->lock);
	root->cnt_guarantee = sectorate;
	root->cnt_unused = sectorate;
	root->cnt_limit = sectorate;
	spin_unlock(&root->lock);

	cki_reset_sectorate(root);
}


/* Callback from ps-iosched.c whenever ps_min/max_sectorates are changed */
int cki_recalc_rootrate(void)
{
	cki_icls_t *rooticls;
	int rc = -1;

	rooticls = ckrm_get_res_class((cki_rcbs.classtype)->default_class, 
				      cki_rcbs.resid, cki_icls_t);
	cki_reset_sectorate(rooticls);
	rc = cki_recalc_abs();
	return rc;
}
	

static void cki_put_psq(cki_icls_t *icls)
{
	struct psdrate *prate;
	struct ckrm_task_class *tskcls;
	
	spin_lock(&icls->lock);
	list_for_each_entry(prate, &icls->rate_list, rate_list) {
		tskcls = container_of(icls->core,struct ckrm_task_class, core);
		if (ps_drop_psq(prate->psd,tskcls)) 
			continue;
		
	}
	spin_unlock(&icls->lock);
}

static void *cki_alloc(struct ckrm_core_class *core,
			 struct ckrm_core_class *parent)
{
	cki_icls_t *icls;
	
	icls = kmalloc(sizeof(cki_icls_t), GFP_ATOMIC);
	if (!icls) {
		printk(KERN_ERR "cki_res_alloc failed GFP_ATOMIC\n");
		return NULL;
	}

	memset(icls, 0, sizeof(cki_icls_t));
	icls->core = core;
	icls->parent = parent;
	spin_lock_init(&icls->lock);

	init_icls_one(icls);

	if (parent == NULL) 
		/* No need to acquire root->share_lock */
		cki_createrootrate(icls, ckid.rootsectorate);
	else
		cki_rates_init(icls);
	
	try_module_get(THIS_MODULE);
	return icls;
}

static void cki_free(void *res)
{
	cki_icls_t *icls = res, *parres, *childres;
	ckrm_core_class_t *child = NULL;
	int maxlimit, resid = cki_rcbs.resid;

	
	if (!res)
		return;

	/* Deallocate CFQ queues */

	/* Currently CFQ queues are deallocated when empty. Since no task 
	 * should belong to this icls, no new requests will get added to the
	 * CFQ queue. 
	 * 
	 * When CFQ switches to persistent queues, call its "put" function
	 * so it gets deallocated after the last pending request is serviced.
	 *
	 */

	parres = ckrm_get_res_class(icls->parent, resid, cki_icls_t);
	if (!parres) 
		return;
	

	/* Update parent's shares */
	spin_lock(&parres->lock);

	child_guarantee_changed(&parres->shares, icls->shares.my_guarantee, 0);
	parres->cnt_unused += icls->cnt_guarantee;

	// run thru parent's children and get the new max_limit of the parent
	ckrm_lock_hier(parres->core);
	maxlimit = 0;
	while ((child = ckrm_get_next_child(parres->core, child)) != NULL) {
		childres = ckrm_get_res_class(child, resid, cki_icls_t);
		if (maxlimit < childres->shares.my_limit) {
			maxlimit = childres->shares.my_limit;
		}
	}
	ckrm_unlock_hier(parres->core);
	if (parres->shares.cur_max_limit < maxlimit) {
		parres->shares.cur_max_limit = maxlimit;
	}
	spin_unlock(&parres->lock);

	/* Drop refcounts on all psq's corresponding to this class */
	cki_put_psq(icls);
	
	cki_rates_del(icls);

	kfree(res);
	module_put(THIS_MODULE);
	return;
}


/* Recalculate absolute shares from relative
 * Caller should hold a lock on icls
 */

static void cki_recalc_propagate(cki_icls_t *res, cki_icls_t *parres)
{

	ckrm_core_class_t *child = NULL;
	cki_icls_t *childres;
	int resid = cki_rcbs.resid;
	u64 temp;

	if (parres) {
		struct ckrm_shares *par = &parres->shares;
		struct ckrm_shares *self = &res->shares;


		if (parres->cnt_guarantee == CKRM_SHARE_DONTCARE) {
			res->cnt_guarantee = CKRM_SHARE_DONTCARE;
		} else if (par->total_guarantee) {
			temp = (u64) self->my_guarantee * 
				parres->cnt_guarantee;
			do_div(temp, par->total_guarantee);
			res->cnt_guarantee = (int) temp;
		} else {
			res->cnt_guarantee = 0;
		}


		if (parres->cnt_limit == CKRM_SHARE_DONTCARE) {
			res->cnt_limit = CKRM_SHARE_DONTCARE;
			__cki_set_sectorate(res,ckid.minsectorate);
		} else {
			if (par->max_limit) {
				temp = (u64) self->my_limit * 
					parres->cnt_limit;
				do_div(temp, par->max_limit);
				res->cnt_limit = (int) temp;
			} else {
				res->cnt_limit = 0;
			}
			__cki_set_sectorate(res,res->cnt_limit);
		}
		
		if (res->cnt_guarantee == CKRM_SHARE_DONTCARE) {
			res->cnt_unused = CKRM_SHARE_DONTCARE;
		} else {
			if (self->total_guarantee) {
				temp = (u64) self->unused_guarantee * 
					res->cnt_guarantee;
				do_div(temp, self->total_guarantee);
				res->cnt_unused = (int) temp;
			} else {
				res->cnt_unused = 0;
			}

		}
		
	}
	// propagate to children
	ckrm_lock_hier(res->core);
	while ((child = ckrm_get_next_child(res->core,child)) != NULL){
		childres = ckrm_get_res_class(child, resid, 
					      cki_icls_t);
		
		spin_lock(&childres->lock);
		cki_recalc_propagate(childres, res);
		spin_unlock(&childres->lock);
	}
	ckrm_unlock_hier(res->core);
}

static int cki_setshare(void *res, struct ckrm_shares *new)
{
	cki_icls_t *icls = res, *parres;
	struct ckrm_shares *cur, *par;
	int rc = -EINVAL, resid = cki_rcbs.resid;

	if (!icls) 
		return rc;

	cur = &icls->shares; 
	if (icls->parent) {
		parres =
		    ckrm_get_res_class(icls->parent, resid, cki_icls_t);
		if (!parres) {
			pr_debug("cki_setshare: invalid resclass\n");
			return -EINVAL;
		}
		spin_lock(&parres->lock);
		spin_lock(&icls->lock);
		par = &parres->shares;
	} else {
		spin_lock(&icls->lock);
		parres = NULL;
		par = NULL;
	}

	rc = set_shares(new, cur, par);

	if ((!rc) && parres) {
		if (parres->cnt_guarantee == CKRM_SHARE_DONTCARE) {
			parres->cnt_unused = CKRM_SHARE_DONTCARE;
		} else if (par->total_guarantee) {
			u64 temp = (u64) par->unused_guarantee * 
				parres->cnt_guarantee;
			do_div(temp, par->total_guarantee);
			parres->cnt_unused = (int) temp;
		} else {
			parres->cnt_unused = 0;
		}
		cki_recalc_propagate(res, parres);
	}
	spin_unlock(&icls->lock);
	if (icls->parent) {
		spin_unlock(&parres->lock);
	}
	return rc;
}

static int cki_getshare(void *res, struct ckrm_shares * shares)
{
	cki_icls_t *icls = res;

	if (!icls)
		return -EINVAL;
	*shares = icls->shares;
	return 0;
}

static int cki_getstats(void *res, struct seq_file *sfile)
{
	cki_icls_t *icls = res;
	struct psdrate *prate;

	if (!icls)
		return -EINVAL;

	seq_printf(sfile, "-----%s controller stats begin -----\n",
		   cki_rcbs.res_name);

	cki_setpsdnames();

	spin_lock(&icls->lock);
	list_for_each_entry(prate, &icls->rate_list, rate_list) {
		seq_printf(sfile,"%s ",prate->psd->name);
		seq_printf(sfile,"used %lu KB/s limit %d KB/s\n",
			   (prate->psrate.navsec * 10)/2,
			   (atomic_read(&(prate->psrate.sectorate))*10)/2);
	}
	spin_unlock(&icls->lock);
	seq_printf(sfile, "-----%s controller stats end -----\n",
		   cki_rcbs.res_name);
	return 0;
}

static int cki_resetstats(void *res)
{
	cki_icls_t *icls = res;

	if (!res)
		return -EINVAL;
	return 0;
}

static void cki_chgcls(void *tsk, void *oldres, void *newres)
{
	/* cki_icls_t *oldicls = oldres, *newicls = newres; */
	
	/* Nothing needs to be done 
	 * Future requests from task will go to the new class's psq
	 * Old ones will continue to get satisfied from the original psq
	 * 
	 */
	return;
}

enum iocfg_token_t {
	ROOTRATE, MINRATE, IOCFGERR
};

/* Token matching for parsing input to this magic file */
static match_table_t iocfg_tokens = {
	{ROOTRATE, "rootsectorate=%d"},
	{MINRATE,"minsectorate=%d"},
	{IOCFGERR, NULL}
};

static int cki_recalc_abs(void)
{
	struct ckrm_core_class *root;
	cki_icls_t *icls;

	root = (cki_rcbs.classtype)->default_class;
	icls = ckrm_get_res_class(root, cki_rcbs.resid, cki_icls_t);
	if (!icls)
		return -EINVAL;

	spin_lock(&icls->lock);
	cki_recalc_propagate(icls, NULL);
	spin_unlock(&icls->lock);
	return 0;
}

	


static int cki_showconfig(void *res, struct seq_file *sfile)
{
	cki_icls_t *icls = res;
	struct cki_data tmp;

	if (!icls)
		return -EINVAL;

	spin_lock(&ckid.cfglock);
	tmp = ckid;
	spin_unlock(&ckid.cfglock);

	seq_printf(sfile,"res=%s,rootsectorate=%d,minsectorate=%d\n",
		   cki_rcbs.res_name,
		   tmp.rootsectorate,
		   tmp.minsectorate);
	return 0;
}
	
static int cki_setconfig(void *res, const char *cfgstr)
{
	char *p, *inpstr = cfgstr;
	int tmp,rc = -EINVAL;
	cki_icls_t *rooticls;


	if (!cfgstr)
		return -EINVAL;
	
	while ((p = strsep(&inpstr, ",")) != NULL) {

		substring_t args[MAX_OPT_ARGS];
		int token;

		
		if (!*p)
			continue;
		
		token = match_token(p, iocfg_tokens, args);
		switch (token) {

		case ROOTRATE: 
			if (match_int(args, &tmp))
				return -EINVAL;

			if (tmp < 0)
				return -EINVAL;

			spin_lock(&(ckid.cfglock));
			ckid.rootsectorate = tmp;
			spin_unlock(&(ckid.cfglock));
			
			rooticls = ckrm_get_res_class(
				(cki_rcbs.classtype)->default_class, 
				cki_rcbs.resid, cki_icls_t);

			cki_setrootrate(rooticls,tmp);
			/* update absolute shares treewide */
			rc = cki_recalc_abs();
			if (rc)
				return rc;
			break;

		case MINRATE:
			if (match_int(args, &tmp))
				return -EINVAL;

			spin_lock(&(ckid.cfglock));
			if (tmp <= 0 || tmp > ckid.rootsectorate) {
				spin_unlock(&(ckid.cfglock));
				return -EINVAL;
			}
			ckid.minsectorate = tmp;
			spin_unlock(&(ckid.cfglock));
			
			/* update absolute shares treewide */
			rc = cki_recalc_abs();
			if (rc)
				return rc;
			break;

		default:
			return -EINVAL;

		}
	}

	return rc;
}



struct ckrm_res_ctlr cki_rcbs = {
	.res_name = "io",
	.res_hdepth = 1,
	.resid = -1,
	.res_alloc = cki_alloc,
	.res_free = cki_free,
	.set_share_values = cki_setshare,
	.get_share_values = cki_getshare,
	.get_stats = cki_getstats,
	.reset_stats = cki_resetstats,
	.show_config = cki_showconfig,
	.set_config = cki_setconfig,
	.change_resclass = cki_chgcls,
};


core_initcall(cki_init);
module_exit(cki_exit);

MODULE_AUTHOR("Shailabh Nagar <nagar@watson.ibm.com>");
MODULE_DESCRIPTION("CKRM Disk I/O Resource Controller");
MODULE_LICENSE("GPL");


