/*
 *  linux/drivers/block/ps-iosched.c
 *
 *  PS, or complete fairness queueing, disk scheduler.
 *
 *  Based on ideas from a previously unfinished io
 *  scheduler (round robin per-process disk scheduling) and Andrea Arcangeli.
 *
 *  Copyright (C) 2003 Jens Axboe <axboe@suse.de>
 */
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/blkdev.h>
#include <linux/elevator.h>
#include <linux/bio.h>
#include <linux/config.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/compiler.h>
#include <linux/hash.h>
#include <linux/rbtree.h>
#include <linux/mempool.h>
#include <linux/kobject.h>
#include <linux/ckrm-io.h>
#include <asm/div64.h>

/*
 * tunables
 */
static int ps_quantum = 4;
static int ps_queued = 8;

#define PS_EPOCH		1000000000
#define PS_HMAX_PCT		80

#define PS_QHASH_SHIFT		6
#define PS_QHASH_ENTRIES	(1 << PS_QHASH_SHIFT)
#define list_entry_qhash(entry)	list_entry((entry), struct ps_queue, ps_hash)

#define PS_MHASH_SHIFT		8
#define PS_MHASH_BLOCK(sec)	((sec) >> 3)
#define PS_MHASH_ENTRIES	(1 << PS_MHASH_SHIFT)
#define PS_MHASH_FN(sec)	(hash_long(PS_MHASH_BLOCK((sec)),PS_MHASH_SHIFT))
#define ON_MHASH(prq)		!list_empty(&(prq)->hash)
#define rq_hash_key(rq)		((rq)->sector + (rq)->nr_sectors)
#define list_entry_hash(ptr)	list_entry((ptr), struct ps_rq, hash)

#define list_entry_psq(ptr)	list_entry((ptr), struct ps_queue, ps_list)

#define RQ_DATA(rq)		((struct ps_rq *) (rq)->elevator_private)

static kmem_cache_t *prq_pool;
static kmem_cache_t *ps_pool;
static mempool_t *ps_mpool;

struct ps_queue {
	struct list_head ps_hash;
	struct list_head ps_list;
	struct rb_root sort_list;
	void *key;
	int queued[2];

	struct ps_data *psd;

	/* limit related settings/stats */
	struct ps_rate *psrate;	

	u64 epstart;		/* current epoch's starting timestamp (ns) */
	u64 epsector[2];	/* Total sectors dispatched in [0] previous
				 * and [1] current epoch
				 */
	unsigned long avsec;	/* avg sectors dispatched/epoch */
	int skipped;		/* queue skipped at last dispatch ? */

	/* Per queue timer to suspend/resume queue from processing */
	struct timer_list timer;
	unsigned long wait_end;
	unsigned long flags;
	struct work_struct work;

#if 0
	/*
	 * with a simple addition like this, we can do io priorities. almost.
	 * does need a split request free list, too.
	 */
	int io_prio
#endif
};

struct ps_rq {
	struct rb_node rb_node;
	sector_t rb_key;

	struct request *request;

	struct ps_queue *ps_queue;

	struct list_head hash;
};

static void ps_put_queue(struct ps_data *psd, struct ps_queue *psq);
static struct ps_queue *ps_find_ps_hash(struct ps_data *psd, void *key);
static void ps_dispatch_sort(struct ps_data *psd, struct ps_queue *psq,
			      struct ps_rq *prq);

extern void print_psdlist(void);

/*
 * lots of deadline iosched dupes, can be abstracted later...
 */
static inline void __ps_del_prq_hash(struct ps_rq *prq)
{
	list_del_init(&prq->hash);
}

static inline void ps_del_prq_hash(struct ps_rq *prq)
{
	if (ON_MHASH(prq))
		__ps_del_prq_hash(prq);
}

static void ps_remove_merge_hints(request_queue_t *q, struct ps_rq *prq)
{
	ps_del_prq_hash(prq);

	if (q->last_merge == prq->request)
		q->last_merge = NULL;
}

static inline void ps_add_prq_hash(struct ps_data *psd, struct ps_rq *prq)
{
	struct request *rq = prq->request;

	BUG_ON(ON_MHASH(prq));

	list_add(&prq->hash, &psd->prq_hash[PS_MHASH_FN(rq_hash_key(rq))]);
}

static struct request *ps_find_rq_hash(struct ps_data *psd, sector_t offset)
{
	struct list_head *hash_list = &psd->prq_hash[PS_MHASH_FN(offset)];
	struct list_head *entry, *next = hash_list->next;

	while ((entry = next) != hash_list) {
		struct ps_rq *prq = list_entry_hash(entry);
		struct request *__rq = prq->request;

		next = entry->next;

		BUG_ON(!ON_MHASH(prq));

		if (!rq_mergeable(__rq)) {
			__ps_del_prq_hash(prq);
			continue;
		}

		if (rq_hash_key(__rq) == offset)
			return __rq;
	}

	return NULL;
}

/*
 * rb tree support functions
 */
#define RB_NONE		(2)
#define RB_EMPTY(node)	((node)->rb_node == NULL)
#define RB_CLEAR(node)	((node)->rb_color = RB_NONE)
#define RB_CLEAR_ROOT(root)	((root)->rb_node = NULL)
#define ON_RB(node)	((node)->rb_color != RB_NONE)
#define rb_entry_prq(node)	rb_entry((node), struct ps_rq, rb_node)
#define rq_rb_key(rq)		(rq)->sector

static inline void ps_del_prq_rb(struct ps_queue *psq, struct ps_rq *prq)
{
	if (ON_RB(&prq->rb_node)) {
		psq->queued[rq_data_dir(prq->request)]--;
		rb_erase(&prq->rb_node, &psq->sort_list);
		prq->ps_queue = NULL;
	}
}

static struct ps_rq *
__ps_add_prq_rb(struct ps_queue *psq, struct ps_rq *prq)
{
	struct rb_node **p = &psq->sort_list.rb_node;
	struct rb_node *parent = NULL;
	struct ps_rq *__prq;

	while (*p) {
		parent = *p;
		__prq = rb_entry_prq(parent);

		if (prq->rb_key < __prq->rb_key)
			p = &(*p)->rb_left;
		else if (prq->rb_key > __prq->rb_key)
			p = &(*p)->rb_right;
		else
			return __prq;
	}

	rb_link_node(&prq->rb_node, parent, p);
	return 0;
}

static void
ps_add_prq_rb(struct ps_data *psd, struct ps_queue *psq,struct ps_rq *prq)
{
	struct request *rq = prq->request;
	struct ps_rq *__alias;

	prq->rb_key = rq_rb_key(rq);
	psq->queued[rq_data_dir(rq)]++;
retry:
	__alias = __ps_add_prq_rb(psq, prq);
	if (!__alias) {
		rb_insert_color(&prq->rb_node, &psq->sort_list);
		prq->ps_queue = psq;
		return;
	}

	ps_dispatch_sort(psd, psq, __alias);
	goto retry;
}

static struct request *
ps_find_rq_rb(struct ps_data *psd, sector_t sector)
{
	struct ps_queue *psq = ps_find_ps_hash(psd, current->taskclass);
	struct rb_node *n;

	if (!psq)
		goto out;

	n = psq->sort_list.rb_node;
	while (n) {
		struct ps_rq *prq = rb_entry_prq(n);

		if (sector < prq->rb_key)
			n = n->rb_left;
		else if (sector > prq->rb_key)
			n = n->rb_right;
		else
			return prq->request;
	}

out:
	return NULL;
}

static void ps_remove_request(request_queue_t *q, struct request *rq)
{
	struct ps_data *psd = q->elevator.elevator_data;
	struct ps_rq *prq = RQ_DATA(rq);

	if (prq) {
		struct ps_queue *psq = prq->ps_queue;

		ps_remove_merge_hints(q, prq);
		list_del_init(&rq->queuelist);

		if (psq) {
			ps_del_prq_rb(psq, prq);

			if (RB_EMPTY(&psq->sort_list))
				ps_put_queue(psd, psq);
		}
	}
}

static int
ps_merge(request_queue_t *q, struct request **req, struct bio *bio)
{
	struct ps_data *psd = q->elevator.elevator_data;
	struct request *__rq;
	int ret;

	ret = elv_try_last_merge(q, bio);
	if (ret != ELEVATOR_NO_MERGE) {
		__rq = q->last_merge;
		goto out_insert;
	}

	__rq = ps_find_rq_hash(psd, bio->bi_sector);
	if (__rq) {
		BUG_ON(__rq->sector + __rq->nr_sectors != bio->bi_sector);

		if (elv_rq_merge_ok(__rq, bio)) {
			ret = ELEVATOR_BACK_MERGE;
			goto out;
		}
	}

	__rq = ps_find_rq_rb(psd, bio->bi_sector + bio_sectors(bio));
	if (__rq) {
		if (elv_rq_merge_ok(__rq, bio)) {
			ret = ELEVATOR_FRONT_MERGE;
			goto out;
		}
	}

	return ELEVATOR_NO_MERGE;
out:
	q->last_merge = __rq;
out_insert:
	*req = __rq;
	return ret;
}

static void ps_merged_request(request_queue_t *q, struct request *req)
{
	struct ps_data *psd = q->elevator.elevator_data;
	struct ps_rq *prq = RQ_DATA(req);

	ps_del_prq_hash(prq);
	ps_add_prq_hash(psd, prq);

	if (ON_RB(&prq->rb_node) && (rq_rb_key(req) != prq->rb_key)) {
		struct ps_queue *psq = prq->ps_queue;

		ps_del_prq_rb(psq, prq);
		ps_add_prq_rb(psd, psq, prq);
	}

	q->last_merge = req;
}

static void
ps_merged_requests(request_queue_t *q, struct request *req,
		    struct request *next)
{
	ps_merged_request(q, req);
	ps_remove_request(q, next);
}

/* Over how many ns is sectorate defined */
#define NS4SCALE  (100000000)

struct ps_rq *dbprq;
struct ps_queue *dbpsq;
unsigned long dbsectorate;

static void __ps_check_limit(struct ps_data *psd,struct ps_queue *psq, int dontskip)
{
	struct ps_rq *prq;
	unsigned long long ts, gap, epoch, tmp;
	unsigned long newavsec, sectorate;

	prq = rb_entry_prq(rb_first(&psq->sort_list));

	dbprq = prq;
	dbpsq = psq;

	ts = sched_clock();
	gap = ts - psq->epstart;
	epoch = psd->ps_epoch;

	sectorate = atomic_read(&psq->psrate->sectorate);
	dbsectorate = sectorate;

	if ((gap >= epoch) || (gap < 0)) {

		if (gap >= (epoch << 1)) {
			psq->epsector[0] = 0;
			psq->epstart = ts ; 
		} else {
			psq->epsector[0] = psq->epsector[1];
			psq->epstart += epoch;
		} 
		psq->epsector[1] = 0;
		gap = ts - psq->epstart;

		tmp  = (psq->epsector[0] + prq->request->nr_sectors) * NS4SCALE;
		do_div(tmp,epoch+gap);

		psq->avsec = (unsigned long)tmp;
		psq->skipped = 0;
		psq->epsector[1] += prq->request->nr_sectors;
		
		psq->psrate->navsec = psq->avsec;
		psq->psrate->sec[0] = psq->epsector[0];
		psq->psrate->sec[1] = psq->epsector[1];
		psq->psrate->timedout++;
		return;
	} else {
		
		tmp = (psq->epsector[0] + psq->epsector[1] + 
		       prq->request->nr_sectors) * NS4SCALE;
		do_div(tmp,epoch+gap);

		newavsec = (unsigned long)tmp;
		if ((newavsec < sectorate) || dontskip) {
			psq->avsec = newavsec ;
			psq->skipped = 0;
			psq->epsector[1] += prq->request->nr_sectors;
			psq->psrate->navsec = psq->avsec;
			psq->psrate->sec[1] = psq->epsector[1];
		} else {
			psq->skipped = 1;
			/* pause q's processing till avsec drops to 
			   ps_hmax_pct % of its value */
			tmp = (epoch+gap) * (100-psd->ps_hmax_pct);
			do_div(tmp,1000000*psd->ps_hmax_pct);
			psq->wait_end = jiffies+ckrm_msecs_to_jiffies(tmp);
		}
	}	
}


static void
ps_dispatch_sort(struct ps_data *psd, struct ps_queue *psq,
		  struct ps_rq *prq)
{
	struct list_head *head = psd->dispatch, *entry = head;
	struct request *__rq;

	ps_del_prq_rb(psq, prq);
	ps_remove_merge_hints(psd->queue, prq);

	if (!list_empty(head)) {
		__rq = list_entry_rq(head->next);

		if (prq->request->sector < __rq->sector) {
			entry = head->prev;
			goto link;
		}
	}

	while ((entry = entry->prev) != head) {
		__rq = list_entry_rq(entry);

		if (prq->request->sector <= __rq->sector)
			break;
	}

link:
	list_add_tail(&prq->request->queuelist, entry);
}

static inline void
__ps_dispatch_requests(request_queue_t *q, struct ps_data *psd,
			struct ps_queue *psq)
{
	struct ps_rq *prq = rb_entry_prq(rb_first(&psq->sort_list));

	ps_dispatch_sort(psd, psq, prq);
}

static int ps_dispatch_requests(request_queue_t *q, struct ps_data *psd)
{
	struct ps_queue *psq;
	struct list_head *entry, *tmp;
	int ret, queued, good_queues;
	int busy_queues, first_round, busy_unlimited;

	if (list_empty(&psd->rr_list))
		return 0;

	queued = ret = 0;
	first_round = 1;
restart:
	good_queues = 0;
	busy_unlimited = 0;
	busy_queues = 0;
	list_for_each_safe(entry, tmp, &psd->rr_list) {
		psq = list_entry_psq(psd->rr_list.next);

		BUG_ON(RB_EMPTY(&psq->sort_list));
		busy_queues++;

		if (first_round || busy_unlimited)
			__ps_check_limit(psd,psq,0);
		else
			__ps_check_limit(psd,psq,1);
		
		if (psq->skipped) {
			psq->psrate->nskip++;
			busy_queues--;
			if (time_before(jiffies, psq->wait_end)) {
				list_del(&psq->ps_list);
				mod_timer(&psq->timer,psq->wait_end);
			}
			continue;
		}
		busy_unlimited++;

		__ps_dispatch_requests(q, psd, psq);

		if (RB_EMPTY(&psq->sort_list)) {
			busy_unlimited--;
			busy_queues--;
			ps_put_queue(psd,psq);
		}

		queued++;
		ret = 1;
	}

	if ((queued < psd->ps_quantum) && (busy_queues || first_round)) {
		first_round = 0;
		goto restart;
	}

	return ret;
}

static struct request *ps_next_request(request_queue_t *q)
{
	struct ps_data *psd = q->elevator.elevator_data;
	struct request *rq;

	if (!list_empty(psd->dispatch)) {
		struct ps_rq *prq;
dispatch:
		rq = list_entry_rq(psd->dispatch->next);

		prq = RQ_DATA(rq);
		if (prq)
			ps_remove_merge_hints(q, prq);

		return rq;
	}

	if (ps_dispatch_requests(q, psd))
		goto dispatch;

	return NULL;
}

void ps_set_sectorate(struct ckrm_core_class *core, int sectorate)
{
	struct ps_data *psd;
	struct ps_queue *psq;
	u64 temp;

	down_read(&psdlistsem);
	list_for_each_entry(psd, &ps_psdlist, psdlist) {
		psq = ps_find_ps_hash(psd,core);
		
		temp = (u64) sectorate * psd->ps_max_sectorate;
		do_div(temp,ckid.rootsectorate);

		atomic_set(&psq->psrate->sectorate, temp);
	}
	up_read(&psdlistsem);
}

static inline struct ps_queue *
__ps_find_ps_hash(struct ps_data *psd, void *key, const ulong hashval)
{
	struct list_head *hash_list = &psd->ps_hash[hashval];
	struct list_head *entry;

	list_for_each(entry, hash_list) {
		struct ps_queue *__psq = list_entry_qhash(entry);

		if (__psq->key == key)
			return __psq;
	}

	return NULL;
}

static struct ps_queue *ps_find_ps_hash(struct ps_data *psd, void *key)
{
	const ulong hashval = hash_ptr(key, PS_QHASH_SHIFT);

	return __ps_find_ps_hash(psd, key, hashval);
}


static void ps_pauseq_timer(unsigned long data)
{
	struct ps_queue *psq = (struct ps_queue *) data;
	kblockd_schedule_work(&psq->work);
}

static void ps_pauseq_work(void *data)
{
	struct ps_queue *psq = (struct ps_queue *) data;
	struct ps_data *psd = psq->psd;
	request_queue_t *q = psd->queue;
	unsigned long flags;
	int emt=0;
	
	spin_lock_irqsave(q->queue_lock, flags);
	list_add_tail(&psq->ps_list,&psd->rr_list);
	psq->skipped = 0;
	if (ps_next_request(q))
		q->request_fn(q);
	else 
	    emt = 1;
	spin_unlock_irqrestore(q->queue_lock, flags);

	if (emt)
	    printk(KERN_WARNING "------> %s adding empty queue to rr_list\n",
		   __FUNCTION__);
}	

static void ps_put_queue(struct ps_data *psd, struct ps_queue *psq)
{
	psd->busy_queues--;
	list_del(&psq->ps_list);
	list_del(&psq->ps_hash);
	mempool_free(psq, ps_mpool);
}

static struct ps_queue *__ps_get_queue(struct ps_data *psd, void *key,
					 int gfp_mask)
{
	const ulong hashval = hash_ptr(current->taskclass, PS_QHASH_SHIFT);
	struct ps_queue *psq, *new_psq = NULL;
	request_queue_t *q = psd->queue;

retry:
	psq = __ps_find_ps_hash(psd, key, hashval);

	if (!psq) {
		if (new_psq) {
			psq = new_psq;
			new_psq = NULL;
		} else if (gfp_mask & __GFP_WAIT) {
			spin_unlock_irq(q->queue_lock);
			new_psq = mempool_alloc(ps_mpool, gfp_mask);
			spin_lock_irq(q->queue_lock);
			goto retry;
		} else
			return NULL;

		memset(psq, 0, sizeof(*psq));

		INIT_LIST_HEAD(&psq->ps_hash);
		INIT_LIST_HEAD(&psq->ps_list);
		RB_CLEAR_ROOT(&psq->sort_list);

		psq->psrate = cki_tsk_psrate(psd,current);
		if (!psq->psrate) {
		    psq->psrate = &cki_def_psrate;
		}

		psq->epstart = sched_clock();
		init_timer(&psq->timer);
		psq->timer.function = ps_pauseq_timer;
		psq->timer.data = (unsigned long) psq;
		INIT_WORK(&psq->work, ps_pauseq_work, psq); 

		psq->key = (void *)current->taskclass;
		psq->queued[0] = psq->queued[1] = 0;
		list_add(&psq->ps_hash, &psd->ps_hash[hashval]);

 		psq->psd = psd;
	}

	if (new_psq)
		mempool_free(new_psq, ps_mpool);

	return psq;
}

static struct ps_queue *ps_get_queue(struct ps_data *psd, void *key,
				       int gfp_mask)
{
	request_queue_t *q = psd->queue;
	struct ps_queue *psq;

	spin_lock_irq(q->queue_lock);
	psq = __ps_get_queue(psd, key, gfp_mask);
	spin_unlock_irq(q->queue_lock);

	return psq;
}

static void ps_enqueue(struct ps_data *psd, struct ps_rq *prq)
{
	struct ps_queue *psq;

	psq = __ps_get_queue(psd, current->taskclass, GFP_ATOMIC);
	if (psq) {
		ps_add_prq_rb(psd, psq, prq);

		if (list_empty(&psq->ps_list)) {
			list_add(&psq->ps_list, &psd->rr_list);
			psd->busy_queues++;
		}
	} else {
		/*
		 * should can only happen if the request wasn't allocated
		 * through blk_alloc_request(), eg stack requests from ide-cd
		 * (those should be removed) _and_ we are in OOM.
		 */
		list_add_tail(&prq->request->queuelist, psd->dispatch);
	}
}

static void
ps_insert_request(request_queue_t *q, struct request *rq, int where)
{
	struct ps_data *psd = q->elevator.elevator_data;
	struct ps_rq *prq = RQ_DATA(rq);

	switch (where) {
		case ELEVATOR_INSERT_BACK:
			while (ps_dispatch_requests(q, psd))
				;
			list_add_tail(&rq->queuelist, psd->dispatch);
			break;
		case ELEVATOR_INSERT_FRONT:
			list_add(&rq->queuelist, psd->dispatch);
			break;
		case ELEVATOR_INSERT_SORT:
			BUG_ON(!blk_fs_request(rq));
			ps_enqueue(psd, prq);
			break;
		default:
			printk("%s: bad insert point %d\n", __FUNCTION__,where);
			return;
	}

	if (rq_mergeable(rq)) {
		ps_add_prq_hash(psd, prq);

		if (!q->last_merge)
			q->last_merge = rq;
	}
}

static int ps_queue_empty(request_queue_t *q)
{
	struct ps_data *psd = q->elevator.elevator_data;

	if (list_empty(psd->dispatch) && list_empty(&psd->rr_list))
		return 1;

	return 0;
}

static struct request *
ps_former_request(request_queue_t *q, struct request *rq)
{
	struct ps_rq *prq = RQ_DATA(rq);
	struct rb_node *rbprev = rb_prev(&prq->rb_node);

	if (rbprev)
		return rb_entry_prq(rbprev)->request;

	return NULL;
}

static struct request *
ps_latter_request(request_queue_t *q, struct request *rq)
{
	struct ps_rq *prq = RQ_DATA(rq);
	struct rb_node *rbnext = rb_next(&prq->rb_node);

	if (rbnext)
		return rb_entry_prq(rbnext)->request;

	return NULL;
}

static int ps_may_queue(request_queue_t *q, int rw)
{
	struct ps_data *psd = q->elevator.elevator_data;
	struct ps_queue *psq;
	int ret = 1;

	if (!psd->busy_queues)
		goto out;

	psq = ps_find_ps_hash(psd, current->taskclass);
	if (psq) {
		int limit = (q->nr_requests - psd->ps_queued) / psd->busy_queues;

		if (limit < 3)
			limit = 3;
		else if (limit > psd->max_queued)
			limit = psd->max_queued;

		if (psq->queued[rw] > limit)
			ret = 0;
	}
out:
	return ret;
}

static void ps_put_request(request_queue_t *q, struct request *rq)
{
	struct ps_data *psd = q->elevator.elevator_data;
	struct ps_rq *prq = RQ_DATA(rq);
	struct request_list *rl;
	int other_rw;

	if (prq) {
		BUG_ON(q->last_merge == rq);
		BUG_ON(ON_MHASH(prq));

		mempool_free(prq, psd->prq_pool);
		rq->elevator_private = NULL;
	}

	/*
	 * work-around for may_queue "bug": if a read gets issued and refused
	 * to queue because writes ate all the allowed slots and no other
	 * reads are pending for this queue, it could get stuck infinitely
	 * since freed_request() only checks the waitqueue for writes when
	 * freeing them. or vice versa for a single write vs many reads.
	 * so check here whether "the other" data direction might be able
	 * to queue and wake them
	 */
	rl = &q->rq;
	other_rw = rq_data_dir(rq) ^ 1;
	if (rl->count[other_rw] <= q->nr_requests) {
		smp_mb();
		if (waitqueue_active(&rl->wait[other_rw]))
			wake_up(&rl->wait[other_rw]);
	}
}

static int ps_set_request(request_queue_t *q, struct request *rq, int gfp_mask)
{
	struct ps_data *psd = q->elevator.elevator_data;
	struct ps_queue *psq;
	struct ps_rq *prq;

	/*
	 * prepare a queue up front, so ps_enqueue() doesn't have to
	 */
	psq = ps_get_queue(psd, current->taskclass, gfp_mask);
	if (!psq)
		return 1;

	prq = mempool_alloc(psd->prq_pool, gfp_mask);
	if (prq) {
		memset(prq, 0, sizeof(*prq));
		RB_CLEAR(&prq->rb_node);
		prq->request = rq;
		prq->ps_queue = NULL;
		INIT_LIST_HEAD(&prq->hash);
		rq->elevator_private = prq;
		return 0;
	}

	return 1;
}

static void ps_put_psd(struct ps_data *psd)
{
	if (!atomic_dec_and_test(&psd->ref))
		return;
	cki_psd_del(psd);
	mempool_destroy(psd->prq_pool);
	kfree(psd->ps_hash);
	kfree(psd->prq_hash);
	if (psd->name)
	    kfree(psd->name);
	cki_psdlist_del(psd);
	kfree(psd);
}


int ps_drop_psq(struct ps_data *psd, void *key)
{
	struct ps_queue *psq = ps_find_ps_hash(psd, key);
	if (!psq)
		return -1;

	ps_put_queue(psd,psq);
	return 0;
}

static void ps_exit(request_queue_t *q, elevator_t *e)
{
	struct ps_data *psd = e->elevator_data;
	e->elevator_data = NULL;

	ps_put_psd(psd);
}

static int ps_init(request_queue_t *q, elevator_t *e)
{
	struct ps_data *psd;
	struct psd_list_entry *psdl;
	int i,rc;

	psd = kmalloc(sizeof(*psd), GFP_KERNEL);
	if (!psd)
		return -ENOMEM;

	memset(psd, 0, sizeof(*psd));
	INIT_LIST_HEAD(&psd->rr_list);

	psdl = kmalloc(sizeof(*psdl), GFP_KERNEL);
	if (!psdl)
		goto out_psd;
	INIT_LIST_HEAD(&psdl->psd_list);
	psdl->psd = psd;

	rc = cki_psd_init(psd);
	if (rc)
		goto out_psdl;

	psd->prq_hash = kmalloc(sizeof(struct list_head) * PS_MHASH_ENTRIES, GFP_KERNEL);
	if (!psd->prq_hash)
		goto out_psdl;

	psd->ps_hash = kmalloc(sizeof(struct list_head) * PS_QHASH_ENTRIES, GFP_KERNEL);
	if (!psd->ps_hash)
		goto out_prqhash;

	psd->prq_pool = mempool_create(BLKDEV_MIN_RQ, mempool_alloc_slab, mempool_free_slab, prq_pool);
	if (!psd->prq_pool)
		goto out_pshash;

	for (i = 0; i < PS_MHASH_ENTRIES; i++)
		INIT_LIST_HEAD(&psd->prq_hash[i]);
	for (i = 0; i < PS_QHASH_ENTRIES; i++)
		INIT_LIST_HEAD(&psd->ps_hash[i]);

	psd->dispatch = &q->queue_head;
	e->elevator_data = psd;
	psd->queue = q;

	cki_psdlist_add(psdl);

	/*
	 * just set it to some high value, we want anyone to be able to queue
	 * some requests. fairness is handled differently
	 */
	psd->max_queued = q->nr_requests;
 	atomic_set(&psd->ref,1);

	psd->ps_queued = ps_queued;
	psd->ps_quantum = ps_quantum;
	psd->ps_epoch = PS_EPOCH;
	psd->ps_hmax_pct = PS_HMAX_PCT;

	return 0;

out_pshash:
	kfree(psd->ps_hash);
out_prqhash:
	kfree(psd->prq_hash);
out_psdl:
	kfree(psdl);
out_psd:
	kfree(psd);
	return -ENOMEM;
}

static int __init ps_slab_setup(void)
{
	prq_pool = kmem_cache_create("prq_pool", sizeof(struct ps_rq), 0, 0,
					NULL, NULL);

	if (!prq_pool)
		panic("ps_iosched: can't init prq pool\n");

	ps_pool = kmem_cache_create("ps_pool", sizeof(struct ps_queue), 0, 0,
					NULL, NULL);

	if (!ps_pool)
		panic("ps_iosched: can't init ps pool\n");

	ps_mpool = mempool_create(64, mempool_alloc_slab, mempool_free_slab, ps_pool);

	if (!ps_mpool)
		panic("ps_iosched: can't init ps mpool\n");

	return 0;
}

subsys_initcall(ps_slab_setup);

/*
 * sysfs parts below -->
 */
struct ps_fs_entry {
	struct attribute attr;
	ssize_t (*show)(struct ps_data *, char *);
	ssize_t (*store)(struct ps_data *, const char *, size_t);
};

static ssize_t
ps_var_show(unsigned int var, char *page)
{
	return sprintf(page, "%d\n", var);
}

static ssize_t
ps_var_store(unsigned int *var, const char *page, size_t count)
{
	char *p = (char *) page;

	*var = simple_strtoul(p, &p, 10);
	return count;
}

#define SHOW_FUNCTION(__FUNC, __VAR)					\
static ssize_t __FUNC(struct ps_data *psd, char *page)		\
{									\
	return ps_var_show(__VAR, (page));				\
}
SHOW_FUNCTION(ps_quantum_show, psd->ps_quantum);
SHOW_FUNCTION(ps_queued_show, psd->ps_queued);
SHOW_FUNCTION(ps_epoch_show, psd->ps_epoch);
SHOW_FUNCTION(ps_hmax_pct_show, psd->ps_hmax_pct);
SHOW_FUNCTION(ps_max_sectorate_show, psd->ps_max_sectorate);
SHOW_FUNCTION(ps_min_sectorate_show, psd->ps_min_sectorate);
#undef SHOW_FUNCTION

#define STORE_FUNCTION(__FUNC, __PTR, MIN, MAX)				\
static ssize_t __FUNC(struct ps_data *psd, const char *page, size_t count)	\
{									\
	int ret = ps_var_store(__PTR, (page), count);			\
	if (*(__PTR) < (MIN))						\
		*(__PTR) = (MIN);					\
	else if (*(__PTR) > (MAX))					\
		*(__PTR) = (MAX);					\
	return ret;							\
}
STORE_FUNCTION(ps_quantum_store, &psd->ps_quantum, 1, INT_MAX);
STORE_FUNCTION(ps_queued_store, &psd->ps_queued, 1, INT_MAX);
STORE_FUNCTION(ps_epoch_store, &psd->ps_epoch, 0, INT_MAX);
STORE_FUNCTION(ps_hmax_pct_store, &psd->ps_hmax_pct, 1, 100);
#undef STORE_FUNCTION

static ssize_t ps_max_sectorate_store(struct ps_data *psd, const char *page, size_t count)
{
	int rc,ret;
	unsigned int hold = psd->ps_max_sectorate;

	ret = ps_var_store(&psd->ps_max_sectorate, (page), count);
	if (psd->ps_max_sectorate < 0)
	    psd->ps_max_sectorate = 0;
	if (psd->ps_max_sectorate > INT_MAX)
	    psd->ps_max_sectorate = INT_MAX;

	rc = cki_recalc_rootrate();
	if (rc) 
		psd->ps_max_sectorate = hold;
	
	return ret;						      
}
    
static ssize_t ps_min_sectorate_store(struct ps_data *psd, const char *page, size_t count)
{
	int rc,ret;
	unsigned int hold = psd->ps_min_sectorate;

	ret = ps_var_store(&psd->ps_min_sectorate, (page), count);
	if (psd->ps_min_sectorate < 0)
	    psd->ps_min_sectorate = 0;
	if (psd->ps_min_sectorate > INT_MAX)
	    psd->ps_min_sectorate = INT_MAX;

	rc = cki_recalc_rootrate();
	if (rc) 
		psd->ps_min_sectorate = hold;
	
	return ret;						      
}


static struct ps_fs_entry ps_quantum_entry = {
	.attr = {.name = "quantum", .mode = S_IRUGO | S_IWUSR },
	.show = ps_quantum_show,
	.store = ps_quantum_store,
};
static struct ps_fs_entry ps_queued_entry = {
	.attr = {.name = "queued", .mode = S_IRUGO | S_IWUSR },
	.show = ps_queued_show,
	.store = ps_queued_store,
};
static struct ps_fs_entry ps_epoch_entry = {
	.attr = {.name = "epoch", .mode = S_IRUGO | S_IWUSR },
	.show = ps_epoch_show,
	.store = ps_epoch_store,
};
static struct ps_fs_entry ps_hmax_pct_entry = {
	.attr = {.name = "hmaxpct", .mode = S_IRUGO | S_IWUSR },
	.show = ps_hmax_pct_show,
	.store = ps_hmax_pct_store,
};
static struct ps_fs_entry ps_max_sectorate_entry = {
	.attr = {.name = "max_sectorate", .mode = S_IRUGO | S_IWUSR },
	.show = ps_max_sectorate_show,
	.store = ps_max_sectorate_store,
};
static struct ps_fs_entry ps_min_sectorate_entry = {
	.attr = {.name = "min_sectorate", .mode = S_IRUGO | S_IWUSR },
	.show = ps_min_sectorate_show,
	.store = ps_min_sectorate_store,
};

static struct attribute *default_attrs[] = {
	&ps_quantum_entry.attr,
	&ps_queued_entry.attr,
	&ps_epoch_entry.attr,
	&ps_hmax_pct_entry.attr,
	&ps_max_sectorate_entry.attr,
	&ps_min_sectorate_entry.attr,
	NULL,
};

#define to_ps(atr) container_of((atr), struct ps_fs_entry, attr)

static ssize_t
ps_attr_show(struct kobject *kobj, struct attribute *attr, char *page)
{
	elevator_t *e = container_of(kobj, elevator_t, kobj);
	struct ps_fs_entry *entry = to_ps(attr);

	if (!entry->show)
		return 0;

	return entry->show(e->elevator_data, page);
}

static ssize_t
ps_attr_store(struct kobject *kobj, struct attribute *attr,
	       const char *page, size_t length)
{
	elevator_t *e = container_of(kobj, elevator_t, kobj);
	struct ps_fs_entry *entry = to_ps(attr);

	if (!entry->store)
		return -EINVAL;

	return entry->store(e->elevator_data, page, length);
}

static struct sysfs_ops ps_sysfs_ops = {
	.show	= ps_attr_show,
	.store	= ps_attr_store,
};

struct kobj_type ps_ktype = {
	.sysfs_ops	= &ps_sysfs_ops,
	.default_attrs	= default_attrs,
};

elevator_t iosched_ps = {
	.elevator_name =		"ps",
	.elevator_ktype =		&ps_ktype,
	.elevator_merge_fn = 		ps_merge,
	.elevator_merged_fn =		ps_merged_request,
	.elevator_merge_req_fn =	ps_merged_requests,
	.elevator_next_req_fn =		ps_next_request,
	.elevator_add_req_fn =		ps_insert_request,
	.elevator_remove_req_fn =	ps_remove_request,
	.elevator_queue_empty_fn =	ps_queue_empty,
	.elevator_former_req_fn =	ps_former_request,
	.elevator_latter_req_fn =	ps_latter_request,
	.elevator_set_req_fn =		ps_set_request,
	.elevator_put_req_fn =		ps_put_request,
	.elevator_may_queue_fn =	ps_may_queue,
	.elevator_init_fn =		ps_init,
	.elevator_exit_fn =		ps_exit,
};

EXPORT_SYMBOL(iosched_ps);
