/* -*- mode: c; c-basic-offset: 8; -*-
 * vim: noexpandtab sw=8 ts=8 sts=0:
 *
 * dlmglue.c
 *
 * Code which implements an OCFS2 specific interface to our DLM.
 *
 * Copyright (C) 2003, 2004 Oracle.  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 021110-1307, USA.
 */

#include <linux/version.h>
#include <linux/types.h>
#include <linux/slab.h>
#include <linux/highmem.h>
#include <linux/mm.h>
#include <linux/smp_lock.h>
#include <linux/crc32.h>
#include <linux/kthread.h>

#include <cluster/heartbeat.h>
#include <cluster/nodemanager.h>
#include <cluster/tcp.h>

#include <dlm/dlmapi.h>

#define MLOG_MASK_PREFIX ML_DLM_GLUE
#include <cluster/masklog.h>

#include "ocfs2.h"

#include "alloc.h"
#include "dlmglue.h"
#include "extent_map.h"
#include "heartbeat.h"
#include "inode.h"
#include "journal.h"
#include "slot_map.h"
#include "uptodate.h"
#include "vote.h"

#include "buffer_head_io.h"


static void ocfs2_inode_ast_func(void *opaque);
static void ocfs2_inode_bast_func(void *opaque,
				  int level);
static void ocfs2_super_ast_func(void *opaque);
static void ocfs2_super_bast_func(void *opaque,
				  int level);
static void ocfs2_rename_ast_func(void *opaque);
static void ocfs2_rename_bast_func(void *opaque,
				   int level);

/* so far, all locks have gotten along with the same unlock ast */
static void ocfs2_unlock_ast_func(void *opaque,
				  dlm_status status);
static int ocfs2_do_unblock_meta(struct inode *inode,
				 int *requeue);
static int ocfs2_unblock_meta(struct ocfs2_lock_res *lockres,
			      int *requeue);
static int ocfs2_unblock_data(struct ocfs2_lock_res *lockres,
			      int *requeue);
static int ocfs2_unblock_osb_lock(struct ocfs2_lock_res *lockres,
				  int *requeue);
typedef void (ocfs2_convert_worker_t)(struct ocfs2_lock_res *, int);
static int ocfs2_generic_unblock_lock(ocfs2_super *osb,
				      struct ocfs2_lock_res *lockres,
				      int *requeue,
				      ocfs2_convert_worker_t *worker);

struct ocfs2_lock_res_ops {
	void (*ast)(void *);
	void (*bast)(void *, int);
	void (*unlock_ast)(void *, dlm_status);
	int  (*unblock)(struct ocfs2_lock_res *, int *);
};

static struct ocfs2_lock_res_ops ocfs2_inode_meta_lops = {
	.ast		= ocfs2_inode_ast_func,
	.bast		= ocfs2_inode_bast_func,
	.unlock_ast	= ocfs2_unlock_ast_func,
	.unblock	= ocfs2_unblock_meta,
};

static void ocfs2_data_convert_worker(struct ocfs2_lock_res *lockres,
				      int blocking);

static struct ocfs2_lock_res_ops ocfs2_inode_data_lops = {
	.ast		= ocfs2_inode_ast_func,
	.bast		= ocfs2_inode_bast_func,
	.unlock_ast	= ocfs2_unlock_ast_func,
	.unblock	= ocfs2_unblock_data,
};

static struct ocfs2_lock_res_ops ocfs2_super_lops = {
	.ast		= ocfs2_super_ast_func,
	.bast		= ocfs2_super_bast_func,
	.unlock_ast	= ocfs2_unlock_ast_func,
	.unblock	= ocfs2_unblock_osb_lock,
};

static struct ocfs2_lock_res_ops ocfs2_rename_lops = {
	.ast		= ocfs2_rename_ast_func,
	.bast		= ocfs2_rename_bast_func,
	.unlock_ast	= ocfs2_unlock_ast_func,
	.unblock	= ocfs2_unblock_osb_lock,
};

static inline int ocfs2_is_inode_lock(struct ocfs2_lock_res *lockres)
{
	return lockres->l_type == OCFS2_LOCK_TYPE_META || 
		lockres->l_type == OCFS2_LOCK_TYPE_DATA;
}

static inline int ocfs2_is_super_lock(struct ocfs2_lock_res *lockres)
{
	return lockres->l_type == OCFS2_LOCK_TYPE_SUPER;
}

static inline int ocfs2_is_rename_lock(struct ocfs2_lock_res *lockres)
{
	return lockres->l_type == OCFS2_LOCK_TYPE_RENAME;
}

static inline ocfs2_super *ocfs2_lock_res_super(struct ocfs2_lock_res *lockres)
{
	BUG_ON(!ocfs2_is_super_lock(lockres) 
	       && !ocfs2_is_rename_lock(lockres));

	return (ocfs2_super *) lockres->l_priv;
}

static inline struct inode *ocfs2_lock_res_inode(struct ocfs2_lock_res *lockres)
{
	OCFS2_ASSERT(ocfs2_is_inode_lock(lockres));

	return (struct inode *) lockres->l_priv;
}

static int ocfs2_lock_create(ocfs2_super *osb,
			     struct ocfs2_lock_res *lockres,
			     int level,
			     int flags);
static inline int ocfs2_may_continue_on_blocked_lock(struct ocfs2_lock_res *lockres,
						     int wanted);
static int ocfs2_cluster_lock(ocfs2_super *osb,
			      struct ocfs2_lock_res *lockres,
			      int level,
			      int lkm_flags,
			      ocfs2_lock_callback cb,
			      unsigned long cb_data);
static void ocfs2_cluster_unlock(ocfs2_super *osb,
				 struct ocfs2_lock_res *lockres,
				 int level);
static inline void ocfs2_generic_handle_downconvert_action(struct ocfs2_lock_res *lockres);
static inline void ocfs2_generic_handle_convert_action(struct ocfs2_lock_res *lockres);
static inline void ocfs2_generic_handle_attach_action(struct ocfs2_lock_res *lockres);
static int ocfs2_generic_handle_bast(struct ocfs2_lock_res *lockres, int level);
static void ocfs2_schedule_blocked_lock(ocfs2_super *osb,
					struct ocfs2_lock_res *lockres);
static void ocfs2_schedule_blocked_inode_lock(struct inode *inode,
					      struct ocfs2_lock_res *lockres);
static inline void ocfs2_recover_from_dlm_error(struct ocfs2_lock_res *lockres,
						int convert);
#define ocfs2_log_dlm_error(_func, _stat, _lockres) do {	\
	mlog(ML_ERROR, "Dlm error \"%s\" while calling %s on "	\
		"resource %s: %s\n", dlm_errname(_stat), _func,	\
		_lockres->l_name, dlm_errmsg(_stat));		\
} while (0)
static void ocfs2_vote_on_unlock(ocfs2_super *osb,
				 struct ocfs2_lock_res *lockres);
/* Called after we refresh our inode, only has any effect if we have
 * an EX lock. This populates the LVB with the initial values for our
 * change set. */
static void ocfs2_reset_meta_lvb_values(struct inode *inode);
static void __ocfs2_stuff_meta_lvb(struct inode *inode);
static void ocfs2_refresh_inode_from_lvb(struct inode *inode);
static void __ocfs2_lvb_on_downconvert(struct ocfs2_lock_res *lockres,
				       int new_level);
static int ocfs2_meta_lock_update(struct inode *inode,
				  struct buffer_head **bh);
static void ocfs2_drop_osb_locks(ocfs2_super *osb);
static inline int ocfs2_highest_compat_lock_level(int level);
static int __ocfs2_downconvert_lock(ocfs2_super *osb,
				    struct ocfs2_lock_res *lockres,
				    int new_level,
				    int lvb);
static int __ocfs2_cancel_convert(ocfs2_super *osb,
				  struct ocfs2_lock_res *lockres);
static inline int ocfs2_can_downconvert_meta_lock(struct inode *inode,
						  struct ocfs2_lock_res *lockres,
						  int new_level);

static inline int ocfs2_lvb_is_trustable(struct ocfs2_lock_res *lockres)
{
	struct ocfs2_lvb *lvb = (struct ocfs2_lvb *) lockres->l_lksb.lvb;
	unsigned int lvb_seq;
	int ret = 0;

	mlog_entry_void();

	spin_lock(&lockres->l_lock);

	lvb_seq = be32_to_cpu(lvb->lvb_seq);
	if (lvb_seq &&
	    lockres->l_local_seq == lvb_seq)
		ret = 1;

	spin_unlock(&lockres->l_lock);

	mlog_exit(ret);
	return ret;
}

static inline void ocfs2_set_local_seq_from_lvb(struct ocfs2_lock_res *lockres)
{
	struct ocfs2_lvb *lvb = (struct ocfs2_lvb *) lockres->l_lksb.lvb;
	unsigned int lvb_seq;

	mlog_entry_void();

	spin_lock(&lockres->l_lock);

	lvb_seq = be32_to_cpu(lvb->lvb_seq);
	if (lvb_seq)
		lockres->l_local_seq = lvb_seq;

	spin_unlock(&lockres->l_lock);

	mlog_exit_void();
}

/* fill in new values as we add them to the lvb. */
static inline void ocfs2_meta_lvb_get_trunc_clusters(struct ocfs2_lock_res *lockres,
						     unsigned int *trunc_clusters)
{
	struct ocfs2_meta_lvb *lvb;

	mlog_entry_void();

	OCFS2_ASSERT(lockres->l_type == OCFS2_LOCK_TYPE_META);

	spin_lock(&lockres->l_lock);
	OCFS2_ASSERT(lockres->l_level > LKM_NLMODE);

	lvb = (struct ocfs2_meta_lvb *) lockres->l_lksb.lvb;
	if (trunc_clusters)
		*trunc_clusters = be32_to_cpu(lvb->lvb_trunc_clusters);

	spin_unlock(&lockres->l_lock);

	mlog_exit_void();
}

static void ocfs2_build_lock_name(enum ocfs2_lock_type type,
				  u64 blkno,
				  u32 generation,
				  char *name)
{
	int len;

	mlog_entry_void();

	BUG_ON(type >= OCFS2_NUM_LOCK_TYPES);

	len = snprintf(name, OCFS2_LOCK_ID_MAX_LEN, "%c%s%016"MLFx64"%08x", 
		       ocfs2_lock_type_char(type), OCFS2_LOCK_ID_PAD, blkno, 
		       generation);

	BUG_ON(len != (OCFS2_LOCK_ID_MAX_LEN - 1));

	mlog(0, "built lock resource with name: %s\n", name);

	mlog_exit_void();
}

static void ocfs2_lock_res_init_common(struct ocfs2_lock_res *res,
				       enum ocfs2_lock_type type,
				       u64 blkno,
				       u32 generation,
				       struct ocfs2_lock_res_ops *ops,
				       void *priv)
{
	ocfs2_build_lock_name(type, blkno, generation, res->l_name);

	res->l_type          = type;
	res->l_ops           = ops;
	res->l_priv          = priv;

	res->l_level         = LKM_IVMODE;
	res->l_requested     = LKM_IVMODE;
	res->l_blocking      = LKM_IVMODE;
	res->l_action        = OCFS2_AST_INVALID;
	res->l_unlock_action = OCFS2_UNLOCK_INVALID;

	res->l_flags         = OCFS2_LOCK_INITIALIZED;
}

void ocfs2_lock_res_init_once(struct ocfs2_lock_res *res)
{
	/* This also clears out the lock status block */
	memset(res, 0, sizeof(struct ocfs2_lock_res));
	spin_lock_init(&res->l_lock);
	init_waitqueue_head(&res->l_event);
	INIT_LIST_HEAD(&res->l_blocked_list);
	INIT_LIST_HEAD(&res->l_flag_cb_list);
}

void ocfs2_inode_lock_res_init(struct ocfs2_lock_res *res,
			       enum ocfs2_lock_type type,
			       struct inode *inode)
{
	struct ocfs2_lock_res_ops *ops;

	BUG_ON(type != OCFS2_LOCK_TYPE_META &&
	       type != OCFS2_LOCK_TYPE_DATA);

	if (type == OCFS2_LOCK_TYPE_META)
		ops = &ocfs2_inode_meta_lops;
	else
		ops = &ocfs2_inode_data_lops;

	ocfs2_lock_res_init_common(res, type, OCFS2_I(inode)->ip_blkno,
				   inode->i_generation, ops, inode);
}

static void ocfs2_super_lock_res_init(struct ocfs2_lock_res *res,
				      ocfs2_super *osb)
{
	/* Superblock lockres doesn't come from a slab so we call init
	 * once on it manually.  */
	ocfs2_lock_res_init_once(res);
	ocfs2_lock_res_init_common(res, OCFS2_LOCK_TYPE_SUPER,
				   OCFS2_SUPER_BLOCK_BLKNO, 0,
				   &ocfs2_super_lops, osb);
}

static void ocfs2_rename_lock_res_init(struct ocfs2_lock_res *res,
				       ocfs2_super *osb)
{
	/* Rename lockres doesn't come from a slab so we call init
	 * once on it manually.  */
	ocfs2_lock_res_init_once(res);
	ocfs2_lock_res_init_common(res, OCFS2_LOCK_TYPE_RENAME, 0, 0,
				   &ocfs2_rename_lops, osb);
}

void ocfs2_lock_res_free(struct ocfs2_lock_res *res)
{
	mlog_entry_void();

	if (!(res->l_flags & OCFS2_LOCK_INITIALIZED))
		return;

	mlog_bug_on_msg(!list_empty(&res->l_blocked_list),
			"Lockres %s is on the blocked list\n",
			res->l_name);
	mlog_bug_on_msg(!list_empty(&res->l_flag_cb_list),
			"Lockres %s has flag callbacks pending\n",
			res->l_name);
	mlog_bug_on_msg(spin_is_locked(&res->l_lock),
			"Lockres %s is locked\n",
			res->l_name);
	mlog_bug_on_msg(res->l_ro_holders,
			"Lockres %s has %u ro holders\n",
			res->l_name, res->l_ro_holders);
	mlog_bug_on_msg(res->l_ex_holders,
			"Lockres %s has %u ex holders\n",
			res->l_name, res->l_ex_holders);

	res->l_local_seq = 0;
	/* Need to clear out the lock status block for the dlm */
	memset(&res->l_lksb, 0, sizeof(res->l_lksb));

	res->l_flags = 0UL;
	mlog_exit_void();
}

static inline void ocfs2_inc_holders(struct ocfs2_lock_res *lockres,
				     int level)
{
	mlog_entry_void();

	OCFS2_ASSERT(lockres);

	switch(level) {
	case LKM_EXMODE:
		lockres->l_ex_holders++;
		break;
	case LKM_PRMODE:
		lockres->l_ro_holders++;
		break;
	default:
		BUG();
	}

	mlog_exit_void();
}

static inline void ocfs2_dec_holders(struct ocfs2_lock_res *lockres,
				     int level)
{
	mlog_entry_void();

	OCFS2_ASSERT(lockres);

	switch(level) {
	case LKM_EXMODE:
		OCFS2_ASSERT(lockres->l_ex_holders);
		lockres->l_ex_holders--;
		break;
	case LKM_PRMODE:
		OCFS2_ASSERT(lockres->l_ro_holders);
		lockres->l_ro_holders--;
		break;
	default:
		BUG();
	}
	mlog_exit_void();
}

/* WARNING: This function lives in a world where the only three lock
 * levels are EX, PR, and NL. It *will* have to be adjusted when more
 * lock types are added. */
static inline int ocfs2_highest_compat_lock_level(int level)
{
	int new_level = LKM_EXMODE;

	if (level == LKM_EXMODE)
		new_level = LKM_NLMODE;
	else if (level == LKM_PRMODE)
		new_level = LKM_PRMODE;
	return new_level;
}

/* XXX must be called with lockres->l_lock held */
static void lockres_set_flags(struct ocfs2_lock_res *lockres, unsigned long newflags)
{
	struct list_head *pos, *tmp;
	struct ocfs2_lockres_flag_callback *fcb;

	assert_spin_locked(&lockres->l_lock);

	lockres->l_flags = newflags;

	list_for_each_safe(pos, tmp, &lockres->l_flag_cb_list) {
		fcb = list_entry(pos, struct ocfs2_lockres_flag_callback,
				 fc_lockres_item);
		if ((lockres->l_flags & fcb->fc_flag_mask) !=
		    fcb->fc_flag_goal)
			continue;

		list_del_init(&fcb->fc_lockres_item);
		fcb->fc_cb(0, fcb->fc_data);
		if (fcb->fc_free_once_called)
			kfree(fcb);
	}
}

static void lockres_or_flags(struct ocfs2_lock_res *lockres, unsigned long or)
{
	lockres_set_flags(lockres, lockres->l_flags | or);
}
static void lockres_clear_flags(struct ocfs2_lock_res *lockres, unsigned long clear)
{
	lockres_set_flags(lockres, lockres->l_flags & ~clear);
}

static inline void ocfs2_generic_handle_downconvert_action(struct ocfs2_lock_res *lockres)
{
	mlog_entry_void();

	OCFS2_ASSERT(lockres->l_flags & OCFS2_LOCK_BUSY);
	OCFS2_ASSERT(lockres->l_flags & OCFS2_LOCK_ATTACHED);
	OCFS2_ASSERT(lockres->l_flags & OCFS2_LOCK_BLOCKED);
	OCFS2_ASSERT(lockres->l_blocking > LKM_NLMODE);

	lockres->l_level = lockres->l_requested;
	if (lockres->l_level <=
	    ocfs2_highest_compat_lock_level(lockres->l_blocking)) {
		lockres->l_blocking = LKM_NLMODE;
		lockres_clear_flags(lockres, OCFS2_LOCK_BLOCKED);
	}
	lockres_clear_flags(lockres, OCFS2_LOCK_BUSY);

	mlog_exit_void();
}

static inline void ocfs2_generic_handle_convert_action(struct ocfs2_lock_res *lockres)
{
	mlog_entry_void();

	OCFS2_ASSERT(lockres->l_flags & OCFS2_LOCK_BUSY);
	OCFS2_ASSERT(lockres->l_flags & OCFS2_LOCK_ATTACHED);

	/* Convert from RO to EX doesn't really need anything as our
	 * information is already up to data. Convert from NL to
	 * *anything* however should mark ourselves as needing an
	 * update */
	if (lockres->l_level == LKM_NLMODE)
		lockres_or_flags(lockres, OCFS2_LOCK_NEEDS_REFRESH);

	lockres->l_level = lockres->l_requested;
	lockres_clear_flags(lockres, OCFS2_LOCK_BUSY);

	mlog_exit_void();
}

static inline void ocfs2_generic_handle_attach_action(struct ocfs2_lock_res *lockres)
{
	mlog_entry_void();

	OCFS2_ASSERT(lockres->l_flags & OCFS2_LOCK_BUSY);
	OCFS2_ASSERT(!(lockres->l_flags & OCFS2_LOCK_ATTACHED));

	if (lockres->l_requested > LKM_NLMODE &&
	    !(lockres->l_flags & OCFS2_LOCK_LOCAL))
		lockres_or_flags(lockres, OCFS2_LOCK_NEEDS_REFRESH);

	lockres->l_level = lockres->l_requested;
	lockres_or_flags(lockres, OCFS2_LOCK_ATTACHED);
	lockres_clear_flags(lockres, OCFS2_LOCK_BUSY);

	mlog_exit_void();
}

static void ocfs2_inode_ast_func(void *opaque)
{
	struct ocfs2_lock_res *lockres = opaque;
	struct inode *inode;
	dlm_lockstatus *lksb;

	mlog_entry_void();

	inode = ocfs2_lock_res_inode(lockres);

	mlog(0, "AST fired for inode %"MLFu64", l_action = %u, type = %s\n",
	     OCFS2_I(inode)->ip_blkno, lockres->l_action,
	     (lockres->l_type == OCFS2_LOCK_TYPE_META) ? "Meta" : "Data");

	OCFS2_ASSERT(ocfs2_is_inode_lock(lockres));

	spin_lock(&lockres->l_lock);

	lksb = &(lockres->l_lksb);
	if (lksb->status != DLM_NORMAL) {
		mlog(ML_ERROR, "ocfs2_inode_ast_func: lksb status value of %u "
			       "on inode %"MLFu64"\n",
		     lksb->status,
		     OCFS2_I(inode)->ip_blkno);
		spin_unlock(&lockres->l_lock);
		mlog_exit_void();
		return;
	}

	switch(lockres->l_action) {
	case OCFS2_AST_ATTACH:
		ocfs2_generic_handle_attach_action(lockres);
		lockres_clear_flags(lockres, OCFS2_LOCK_LOCAL);
		break;
	case OCFS2_AST_CONVERT:
		ocfs2_generic_handle_convert_action(lockres);
		break;
	case OCFS2_AST_DOWNCONVERT:
		ocfs2_generic_handle_downconvert_action(lockres);
		break;
	default:
		mlog(ML_ERROR, "lockres %s: ast fired with invalid action: %u "
		     "lockres flags = 0x%lx, unlock action: %u\n",
		     lockres->l_name, lockres->l_action, lockres->l_flags,
		     lockres->l_unlock_action);

		BUG();
	}

	/* data locking ignores refresh flag for now. */
	if (lockres->l_type == OCFS2_LOCK_TYPE_DATA)
		lockres_clear_flags(lockres, OCFS2_LOCK_NEEDS_REFRESH);

	/* set it to something invalid so if we get called again we
	 * can catch it. */
	lockres->l_action = OCFS2_AST_INVALID;
	spin_unlock(&lockres->l_lock);
	wake_up(&lockres->l_event);

	mlog_exit_void();
}

static int ocfs2_generic_handle_bast(struct ocfs2_lock_res *lockres,
				     int level)
{
	int needs_downconvert = 0;
	mlog_entry_void();

	assert_spin_locked(&lockres->l_lock);

	lockres_or_flags(lockres, OCFS2_LOCK_BLOCKED);

	if (level > lockres->l_blocking) {
		/* only schedule a downconvert if we haven't already scheduled
		 * one that goes low enough to satisfy the level we're 
		 * blocking.  this also catches the case where we get 
		 * duplicate BASTs */
		if (ocfs2_highest_compat_lock_level(level) <
		    ocfs2_highest_compat_lock_level(lockres->l_blocking))
			needs_downconvert = 1;

		lockres->l_blocking = level;
	}

	mlog_exit(needs_downconvert);
	return needs_downconvert;
}

static void ocfs2_inode_bast_func(void *opaque, int level)
{
	struct ocfs2_lock_res *lockres = opaque;
	struct inode *inode;
	ocfs2_super *osb;
	int needs_downconvert;

	mlog_entry_void();

	OCFS2_ASSERT(ocfs2_is_inode_lock(lockres));

	inode = ocfs2_lock_res_inode(lockres);
	osb = OCFS2_SB(inode->i_sb);

	mlog(0, "BAST fired for inode %"MLFu64", blocking = %d, level = %d "
	     "type = %s\n", OCFS2_I(inode)->ip_blkno, level,
	     lockres->l_level, 
	     (lockres->l_type == OCFS2_LOCK_TYPE_META) ? "Meta" : "Data");

	OCFS2_ASSERT(level > LKM_NLMODE);

	spin_lock(&lockres->l_lock);

	needs_downconvert = ocfs2_generic_handle_bast(lockres, level);
	if (needs_downconvert)
		ocfs2_schedule_blocked_inode_lock(inode, lockres);

	spin_unlock(&lockres->l_lock);

	ocfs2_kick_vote_thread(osb);
	/* TODO: Is a wake_up call here really necessary? */
	wake_up(&lockres->l_event);

	mlog_exit_void();
}

static void ocfs2_generic_ast_func(struct ocfs2_lock_res *lockres)
{
	dlm_lockstatus *lksb = &lockres->l_lksb;

	spin_lock(&lockres->l_lock);

	if (lksb->status != DLM_NORMAL) {
		mlog(ML_ERROR, "lockres %s: lksb status value of %u!\n",
		     lockres->l_name, lksb->status);
		spin_unlock(&lockres->l_lock);
		return;
	}

	switch(lockres->l_action) {
	case OCFS2_AST_ATTACH:
		ocfs2_generic_handle_attach_action(lockres);
		break;
	case OCFS2_AST_CONVERT:
		ocfs2_generic_handle_convert_action(lockres);
		break;
	case OCFS2_AST_DOWNCONVERT:
		ocfs2_generic_handle_downconvert_action(lockres);
		break;
	default:
		BUG();
	}

	/* set it to something invalid so if we get called again we
	 * can catch it. */
	lockres->l_action = OCFS2_AST_INVALID;
	spin_unlock(&lockres->l_lock);

	wake_up(&lockres->l_event);
}

static void ocfs2_generic_bast_func(ocfs2_super *osb,
				    struct ocfs2_lock_res *lockres,
				    int level,
				    int ignore_refresh)
{
	int needs_downconvert;

	mlog_entry_void();

	spin_lock(&lockres->l_lock);
	needs_downconvert = ocfs2_generic_handle_bast(lockres, level);
	if (needs_downconvert)
		ocfs2_schedule_blocked_lock(osb, lockres);

	if (ignore_refresh)
		lockres_clear_flags(lockres, OCFS2_LOCK_NEEDS_REFRESH);

	spin_unlock(&lockres->l_lock);

	ocfs2_kick_vote_thread(osb);

	wake_up(&lockres->l_event);
	mlog_exit_void();
}

static void ocfs2_super_ast_func(void *opaque)
{
	struct ocfs2_lock_res *lockres = opaque;

	mlog_entry_void();
	mlog(0, "Superblock AST fired\n");

	BUG_ON(!ocfs2_is_super_lock(lockres));
	ocfs2_generic_ast_func(lockres);

	mlog_exit_void();
}

static void ocfs2_super_bast_func(void *opaque,
				  int level)
{
	struct ocfs2_lock_res *lockres = opaque;
	ocfs2_super *osb;

	mlog_entry_void();
	mlog(0, "Superblock BAST fired\n");

	BUG_ON(!ocfs2_is_super_lock(lockres));
       	osb = ocfs2_lock_res_super(lockres);
	ocfs2_generic_bast_func(osb, lockres, level, 0);

	mlog_exit_void();
}

static void ocfs2_rename_ast_func(void *opaque)
{
	struct ocfs2_lock_res *lockres = opaque;

	mlog_entry_void();

	mlog(0, "Rename AST fired\n");

	BUG_ON(!ocfs2_is_rename_lock(lockres));

	ocfs2_generic_ast_func(lockres);

	mlog_exit_void();
}

static void ocfs2_rename_bast_func(void *opaque,
				   int level)
{
	struct ocfs2_lock_res *lockres = opaque;
	ocfs2_super *osb;

	mlog_entry_void();

	mlog(0, "Rename BAST fired\n");

	BUG_ON(!ocfs2_is_rename_lock(lockres));

	osb = ocfs2_lock_res_super(lockres);
	ocfs2_generic_bast_func(osb, lockres, level, 1);

	mlog_exit_void();
}

static inline void ocfs2_recover_from_dlm_error(struct ocfs2_lock_res *lockres,
						int convert)
{
	mlog_entry_void();
	spin_lock(&lockres->l_lock);
	lockres_clear_flags(lockres, OCFS2_LOCK_BUSY);
	if (convert)
		lockres->l_action = OCFS2_AST_INVALID;
	else
		lockres->l_unlock_action = OCFS2_UNLOCK_INVALID;
	spin_unlock(&lockres->l_lock);

	wake_up(&lockres->l_event);
	mlog_exit_void();
}

/* Note: If we detect another process working on the lock (i.e.,
 * OCFS2_LOCK_BUSY), we'll bail out returning 0. It's up to the caller
 * to do the right thing in that case. 
 */
static int ocfs2_lock_create(ocfs2_super *osb,
			     struct ocfs2_lock_res *lockres,
			     int level,
			     int flags)
{
	int ret = 0;
	dlm_status status;

	mlog_entry_void();

	mlog(0, "lock %s, level = %d, flags = %d\n", lockres->l_name, level, 
	     flags);

	spin_lock(&lockres->l_lock);
	if ((lockres->l_flags & OCFS2_LOCK_ATTACHED) || 
	    (lockres->l_flags & OCFS2_LOCK_BUSY)) {
		spin_unlock(&lockres->l_lock);
		goto bail;
	}

	lockres->l_action = OCFS2_AST_ATTACH;
	lockres->l_requested = level;
	lockres_or_flags(lockres, OCFS2_LOCK_BUSY);
	spin_unlock(&lockres->l_lock);

	status = dlmlock(osb->dlm,
			 level,
			 &lockres->l_lksb,
			 flags,
			 lockres->l_name,
			 lockres->l_ops->ast,
			 lockres,
			 lockres->l_ops->bast);
	if (status != DLM_NORMAL) {
		ocfs2_log_dlm_error("dlmlock", status, lockres);
		ret = -EINVAL;
		ocfs2_recover_from_dlm_error(lockres, 1);
	}

	mlog(0, "lock %s, successfull return from dlmlock\n", lockres->l_name);

bail:
	mlog_exit(ret);
	return ret;
}

static inline int ocfs2_check_wait_flag(struct ocfs2_lock_res *lockres,
					int flag)
{
	int ret;

	spin_lock(&lockres->l_lock);
	ret = lockres->l_flags & flag;
	spin_unlock(&lockres->l_lock);

	return ret;
}

static inline void ocfs2_wait_on_busy_lock(struct ocfs2_lock_res *lockres)

{
	wait_event(lockres->l_event,
		   !ocfs2_check_wait_flag(lockres, OCFS2_LOCK_BUSY));
}

static inline void ocfs2_wait_on_blocked_lock(struct ocfs2_lock_res *lockres)

{
	wait_event(lockres->l_event,
		   !ocfs2_check_wait_flag(lockres, OCFS2_LOCK_BLOCKED));
}

static inline void ocfs2_wait_on_refreshing_lock(struct ocfs2_lock_res *lockres)

{
	wait_event(lockres->l_event,
		   !ocfs2_check_wait_flag(lockres, OCFS2_LOCK_REFRESHING));
}

static void lockres_add_flag_callback(struct ocfs2_lock_res *lockres,
				      struct ocfs2_lockres_flag_callback *fcb,
				      unsigned long mask, unsigned long goal)
{
	BUG_ON(!list_empty(&fcb->fc_lockres_item));
	BUG_ON(fcb->fc_cb == NULL);

	assert_spin_locked(&lockres->l_lock);

	list_add_tail(&fcb->fc_lockres_item, &lockres->l_flag_cb_list);
	fcb->fc_flag_mask = mask;
	fcb->fc_flag_goal = goal;
}

/* predict what lock level we'll be dropping down to on behalf
 * of another node, and return true if the currently wanted
 * level will be compatible with it. */
static inline int ocfs2_may_continue_on_blocked_lock(struct ocfs2_lock_res *lockres,
						     int wanted)
{
	OCFS2_ASSERT(lockres->l_flags & OCFS2_LOCK_BLOCKED);

	return wanted <= ocfs2_highest_compat_lock_level(lockres->l_blocking);
}

/* these are generic and could be used elsewhere */
struct ocfs2_status_completion {
	int			sc_status;
	struct completion	sc_complete;
};

static void ocfs2_status_completion_cb(int rc, unsigned long data)
{
	struct ocfs2_status_completion *sc;

	sc = (struct ocfs2_status_completion *)data;
	sc->sc_status = rc;
	complete(&sc->sc_complete);
}

static int ocfs2_cluster_lock(ocfs2_super *osb,
			      struct ocfs2_lock_res *lockres,
			      int level,
			      int lkm_flags,
			      ocfs2_lock_callback cb,
			      unsigned long cb_data)
{
	struct ocfs2_lockres_flag_callback _fcb, *fcb = &_fcb;
	struct ocfs2_status_completion sc;
	dlm_status status;
	int ret;
	int catch_signals = 1;

	mlog_entry_void();

	if (cb != NULL) {
		fcb = kmalloc(sizeof(*fcb), GFP_NOFS);
		if (fcb == NULL) {
			ret = -ENOMEM;
			goto out;
		}
		fcb->fc_cb = cb;
		fcb->fc_data = cb_data;
		fcb->fc_free_once_called = 1;
	} else {
		init_completion(&sc.sc_complete);
		fcb->fc_cb = ocfs2_status_completion_cb;
		fcb->fc_data = (unsigned long)&sc;
		fcb->fc_free_once_called = 0;
	}

	INIT_LIST_HEAD(&fcb->fc_lockres_item);

again:
	if (catch_signals && signal_pending(current)) {
		ret = -ERESTARTSYS;
		goto out;
	}

	spin_lock(&lockres->l_lock);

	/* We only compare against the currently granted level
	 * here. If the lock is blocked waiting on a downconvert,
	 * we'll get caught below. */
	if (lockres->l_flags & OCFS2_LOCK_BUSY &&
	    level > lockres->l_level) {
		/* is someone sitting in dlm_lock? If so, wait on
		 * them. */
		lockres_add_flag_callback(lockres, fcb, OCFS2_LOCK_BUSY, 0);
		ret = -EIOCBQUEUED;
		goto unlock;
	}

	if (!(lockres->l_flags & OCFS2_LOCK_ATTACHED)) {
		/* lock has not been created yet. */
		spin_unlock(&lockres->l_lock);

		ret = ocfs2_lock_create(osb, lockres, LKM_NLMODE, 0);
		if (ret < 0) {
			mlog_errno(ret);
			goto out;
		}
		goto again;
	}

	if (lockres->l_flags & OCFS2_LOCK_BLOCKED &&
	    !ocfs2_may_continue_on_blocked_lock(lockres, level)) {
		/* is the lock is currently blocked on behalf of
		 * another node */
		lockres_add_flag_callback(lockres, fcb, OCFS2_LOCK_BLOCKED, 0);
		ret = -EIOCBQUEUED;
		goto unlock;
	}

	if (level > lockres->l_level) {
		if (lockres->l_action != OCFS2_AST_INVALID)
			mlog(ML_ERROR, "lockres %s has action %u pending\n",
			     lockres->l_name, lockres->l_action);

		lockres->l_action = OCFS2_AST_CONVERT;
		lockres->l_requested = level;
		lockres_or_flags(lockres, OCFS2_LOCK_BUSY);
		spin_unlock(&lockres->l_lock);

		OCFS2_ASSERT(level != LKM_IVMODE);
		OCFS2_ASSERT(level != LKM_NLMODE);

		mlog(0, "lock %s, convert from %d to level = %d\n",
		     lockres->l_name, lockres->l_level, level);

		/* call dlm_lock to upgrade lock now */
		status = dlmlock(osb->dlm,
				 level,
				 &lockres->l_lksb,
				 lkm_flags|LKM_CONVERT|LKM_VALBLK,
				 lockres->l_name,
				 lockres->l_ops->ast,
				 lockres,
				 lockres->l_ops->bast);
		if (status != DLM_NORMAL) {
			if ((lkm_flags & LKM_NOQUEUE) &&
			    (status == DLM_NOTQUEUED))
				ret = -EAGAIN;
			else {
				ocfs2_log_dlm_error("dlmlock", status,
						    lockres);
				ret = -EINVAL;
			}
			ocfs2_recover_from_dlm_error(lockres, 1);
			goto out;
		}

		mlog(0, "lock %s, successfull return from dlmlock\n",
		     lockres->l_name);

		/* At this point we've gone inside the dlm and need to
		 * complete our work regardless. */
		catch_signals = 0;

		/* wait for busy to clear and carry on */
		goto again;
	}

	/* Ok, if we get here then we're good to go. */
	ocfs2_inc_holders(lockres, level);

	ret = 0;
unlock:
	spin_unlock(&lockres->l_lock);
out:
	if (ret == -EIOCBQUEUED && fcb->fc_cb == ocfs2_status_completion_cb) {
		wait_for_completion(&sc.sc_complete);
		ret = sc.sc_status;
		if (ret == 0)
			goto again;
	}

	if (ret && fcb != NULL && fcb != &_fcb)
		kfree(fcb);

	mlog_exit(ret);
	return ret;
}

static void ocfs2_cluster_unlock(ocfs2_super *osb,
				 struct ocfs2_lock_res *lockres,
				 int level)
{
	mlog_entry_void();
	spin_lock(&lockres->l_lock);
	ocfs2_dec_holders(lockres, level);
	ocfs2_vote_on_unlock(osb, lockres);
	spin_unlock(&lockres->l_lock);
	mlog_exit_void();
}

/* Grants us an EX lock on the data and metadata resources, skipping
 * the normal cluster directory lookup. Use this ONLY on newly created
 * inodes which other nodes can't possibly see, and which haven't been
 * hashed in the inode hash yet. This can give us a good performance
 * increase as it'll skip the network broadcast normally associated
 * with creating a new lock resource. */
int ocfs2_create_new_inode_locks(struct inode *inode)
{
	int status;
	ocfs2_super *osb = OCFS2_SB(inode->i_sb);
	struct ocfs2_lock_res *lockres;

	OCFS2_ASSERT(inode);
	OCFS2_ASSERT(ocfs2_inode_is_new(inode));

	mlog_entry_void();

	mlog(0, "Inode %"MLFu64"\n", OCFS2_I(inode)->ip_blkno);

	/* NOTE: That we don't increment any of the holder counts, nor
	 * do we add anything to a journal handle. Since this is
	 * supposed to be a new inode which the cluster doesn't know
	 * about yet, there is no need to.  As far as the LVB handling
	 * is concerned, this is basically like acquiring an EX lock
	 * on a resource which has an invalid one -- we'll set it
	 * valid when we release the EX. */

	lockres = &OCFS2_I(inode)->ip_meta_lockres;

	spin_lock(&lockres->l_lock);
	OCFS2_ASSERT(!(lockres->l_flags & OCFS2_LOCK_ATTACHED));
	lockres_or_flags(lockres, OCFS2_LOCK_LOCAL);
	spin_unlock(&lockres->l_lock);

	status = ocfs2_lock_create(osb, lockres, LKM_EXMODE, LKM_LOCAL);
	if (status < 0) {
		mlog_errno(status);
		goto bail;
	}

	lockres = &OCFS2_I(inode)->ip_data_lockres;

	spin_lock(&lockres->l_lock);
	OCFS2_ASSERT(!(lockres->l_flags & OCFS2_LOCK_ATTACHED));
	lockres_or_flags(lockres, OCFS2_LOCK_LOCAL);
	spin_unlock(&lockres->l_lock);

	status = ocfs2_lock_create(osb, lockres, LKM_EXMODE, LKM_LOCAL);
	if (status < 0) {
		mlog_errno(status);
		goto bail;
	}

	status = 0;
bail:
	mlog_exit(status);
	return status;
}

int ocfs2_data_lock(struct inode *inode,
		    int write)
{
	int status, level;
	struct ocfs2_lock_res *lockres;

	OCFS2_ASSERT(inode);

	mlog_entry_void();

	mlog(0, "inode %"MLFu64" take %s DATA lock\n",
	     OCFS2_I(inode)->ip_blkno,
	     write ? "EXMODE" : "PRMODE");

	lockres = &OCFS2_I(inode)->ip_data_lockres;

	level = write ? LKM_EXMODE : LKM_PRMODE;

	status = ocfs2_cluster_lock(OCFS2_SB(inode->i_sb), lockres, level, 0,
				    NULL, 0);
	if (status < 0)
		mlog_errno(status);

	mlog_exit(status);
	return status;
}

static void ocfs2_vote_on_unlock(ocfs2_super *osb,
				 struct ocfs2_lock_res *lockres)
{
	int kick = 0;

	mlog_entry_void();

	/* If we know that another node is waiting on our lock, kick
	 * the vote thread * pre-emptively when we reach a release
	 * condition. */
	if (lockres->l_flags & OCFS2_LOCK_BLOCKED) {
		switch(lockres->l_blocking) {
		case LKM_EXMODE:
			if (!lockres->l_ex_holders && !lockres->l_ro_holders)
				kick = 1;
			break;
		case LKM_PRMODE:
			if (!lockres->l_ex_holders)
				kick = 1;
			break;
		default:
			BUG();
		}
	}

	if (kick)
		ocfs2_kick_vote_thread(osb);

	mlog_exit_void();
}

void ocfs2_data_unlock(struct inode *inode,
		       int write)
{
	int level = write ? LKM_EXMODE : LKM_PRMODE;
	struct ocfs2_lock_res *lockres = &OCFS2_I(inode)->ip_data_lockres;

	mlog_entry_void();

	mlog(0, "inode %"MLFu64" drop %s DATA lock\n",
	     OCFS2_I(inode)->ip_blkno, 
	     write ? "EXMODE" : "PRMODE");

	ocfs2_cluster_unlock(OCFS2_SB(inode->i_sb), lockres, level);

	mlog_exit_void();
}

#define OCFS2_SEC_SHIFT  34
#define OCFS2_NSEC_MASK  ((1ULL << (64 - OCFS2_SEC_SHIFT)) - 1)

/* LVB only has room for 64 bits of time here so we pack it for
 * now. */
static u64 ocfs2_pack_timespec(struct timespec *spec)
{
	u64 res;
	u64 sec = spec->tv_sec;
	u32 nsec = spec->tv_nsec;

	res = (sec << OCFS2_SEC_SHIFT) | (nsec & OCFS2_NSEC_MASK);

	return res;
}

/* Call this with the lockres locked. I am reasonably sure we don't
 * need ip_lock in this function as anyone who would be changing those
 * values is supposed to be blocked in ocfs2_meta_lock right now.
 * This doesn't stuff trunc_clusters as that should already be up to
 * date from the truncate code. */
static void __ocfs2_stuff_meta_lvb(struct inode *inode)
{
	unsigned int isize_off;
	struct ocfs2_inode_info *oi = OCFS2_I(inode);
	struct ocfs2_lock_res *lockres = &oi->ip_meta_lockres;
	struct ocfs2_meta_lvb *lvb;

	mlog_entry_void();

	lvb = (struct ocfs2_meta_lvb *) lockres->l_lksb.lvb;

	/* take advantage of the fact that we don't support holes --
	 * we only need the byte offset from i_clusters which nicely
	 * cuts our lvb storage requirement to a u32 */
	isize_off = ocfs2_clusters_to_bytes(inode->i_sb, oi->ip_clusters) 
		- i_size_read(inode);

	lvb->lvb_isize_off = cpu_to_be32(isize_off);
	lvb->lvb_iclusters = cpu_to_be32(oi->ip_clusters);
	lvb->lvb_iuid      = cpu_to_be32(inode->i_uid);
	lvb->lvb_igid      = cpu_to_be32(inode->i_gid);
	lvb->lvb_imode     = cpu_to_be16(inode->i_mode);
	lvb->lvb_inlink    = cpu_to_be16(inode->i_nlink);
	lvb->lvb_iatime_packed  =
		cpu_to_be64(ocfs2_pack_timespec(&inode->i_atime));
	lvb->lvb_ictime_packed =
		cpu_to_be64(ocfs2_pack_timespec(&inode->i_ctime));
	lvb->lvb_imtime_packed =
		cpu_to_be64(ocfs2_pack_timespec(&inode->i_mtime));

	mlog_exit_void();
}

static void ocfs2_unpack_timespec(struct timespec *spec,
				  u64 packed_time)
{
	spec->tv_sec = packed_time >> OCFS2_SEC_SHIFT;
	spec->tv_nsec = packed_time & OCFS2_NSEC_MASK;
}

static void ocfs2_refresh_inode_from_lvb(struct inode *inode)
{
	loff_t isize;
	struct ocfs2_inode_info *oi = OCFS2_I(inode);
	struct ocfs2_lock_res *lockres = &oi->ip_meta_lockres;
	struct ocfs2_meta_lvb *lvb;

	mlog_entry_void();

	lvb = (struct ocfs2_meta_lvb *) lockres->l_lksb.lvb;

	mlog(0, "seq %u, trunc clusters %u, iclusters %u\n",
	     be32_to_cpu(lvb->lvb.lvb_seq),
	     be32_to_cpu(lvb->lvb_trunc_clusters),
	     be32_to_cpu(lvb->lvb_iclusters));
	mlog(0, "uid %u, gid %u, isize offset %u, mode 0x%x\n",
	     be32_to_cpu(lvb->lvb_iuid), be32_to_cpu(lvb->lvb_igid),
	     be32_to_cpu(lvb->lvb_isize_off), be16_to_cpu(lvb->lvb_imode));
	mlog(0, "nlink %u, atime_packed %"MLFu64", ctime_packed %"MLFx64", "
	     "mtime_packed %"MLFx64"\n",
	     be16_to_cpu(lvb->lvb_inlink), be64_to_cpu(lvb->lvb_iatime_packed),
	     be64_to_cpu(lvb->lvb_ictime_packed),
	     be64_to_cpu(lvb->lvb_imtime_packed));

	/* We're safe here without the lockres lock... */
	spin_lock(&oi->ip_lock);
	oi->ip_clusters = be32_to_cpu(lvb->lvb_iclusters);

	isize = ocfs2_clusters_to_bytes(inode->i_sb, oi->ip_clusters)
		- be32_to_cpu(lvb->lvb_isize_off);
	i_size_write(inode, isize);

	/* fast-symlinks are a special case */
	if (S_ISLNK(inode->i_mode) && !oi->ip_clusters)
		inode->i_blocks = 0;
	else
		inode->i_blocks =
			ocfs2_align_bytes_to_sectors(i_size_read(inode));

	inode->i_uid     = be32_to_cpu(lvb->lvb_iuid);
	inode->i_gid     = be32_to_cpu(lvb->lvb_igid);
	inode->i_mode    = be16_to_cpu(lvb->lvb_imode);
	inode->i_nlink   = be16_to_cpu(lvb->lvb_inlink);
	ocfs2_unpack_timespec(&inode->i_atime,
			      be64_to_cpu(lvb->lvb_iatime_packed));
	ocfs2_unpack_timespec(&inode->i_mtime,
			      be64_to_cpu(lvb->lvb_imtime_packed));
	ocfs2_unpack_timespec(&inode->i_ctime,
			      be64_to_cpu(lvb->lvb_ictime_packed));
	spin_unlock(&oi->ip_lock);

	mlog_exit_void();
}

static void ocfs2_reset_meta_lvb_values(struct inode *inode)
{
	struct ocfs2_lock_res *lockres = &OCFS2_I(inode)->ip_meta_lockres;
	struct ocfs2_meta_lvb *lvb;
	u32 i_clusters;

	mlog_entry_void();

	lvb = (struct ocfs2_meta_lvb *) lockres->l_lksb.lvb;

	spin_lock(&OCFS2_I(inode)->ip_lock);
	i_clusters = OCFS2_I(inode)->ip_clusters;
	spin_unlock(&OCFS2_I(inode)->ip_lock);

	spin_lock(&lockres->l_lock);
	if (lockres->l_level == LKM_EXMODE)
		lvb->lvb_trunc_clusters = cpu_to_be32(i_clusters);
	spin_unlock(&lockres->l_lock);

	mlog_exit_void();
}

static void __ocfs2_lvb_on_downconvert(struct ocfs2_lock_res *lockres,
				       int new_level)
{
	struct ocfs2_lvb *lvb = (struct ocfs2_lvb *) lockres->l_lksb.lvb;
	u32 lvb_seq = be32_to_cpu(lvb->lvb_seq);

	mlog_entry_void();

	if (lockres->l_level == LKM_EXMODE) {
		lvb_seq++;
		/* Overflow? */
		if (!lvb_seq)
			lvb_seq = 1;
		lockres->l_local_seq = lvb_seq;
		if (new_level == LKM_NLMODE)
			lockres->l_local_seq++;
		lvb->lvb_seq = cpu_to_be32(lvb_seq);
	} else if (lockres->l_level == LKM_PRMODE) {
		if (lvb_seq)
			lockres->l_local_seq++;
	}

	mlog_exit_void();
}

/* Determine whether a lock resource needs to be refreshed, and
 * arbitrate who gets to refresh it.
 *
 *   0 means no refresh needed.
 *
 *   > 0 means you need to refresh this and you MUST call
 *   ocfs2_complete_lock_res_refresh afterwards. */
static int ocfs2_should_refresh_lock_res(struct ocfs2_lock_res *lockres)
{

	int status = 0;
	mlog_entry_void();

refresh_check:
	spin_lock(&lockres->l_lock);
	if (!(lockres->l_flags & OCFS2_LOCK_NEEDS_REFRESH)) {
		spin_unlock(&lockres->l_lock);
		goto bail;
	}

	if (lockres->l_flags & OCFS2_LOCK_REFRESHING) {
		spin_unlock(&lockres->l_lock);

		ocfs2_wait_on_refreshing_lock(lockres);
		goto refresh_check;
	}

	/* Ok, I'll be the one to refresh this lock. */
	lockres_or_flags(lockres, OCFS2_LOCK_REFRESHING);
	spin_unlock(&lockres->l_lock);

	status = 1;
bail:
	mlog_exit(status);
	return status;
}

/* If status is non zero, I'll mark it as not being in refresh
 * anymroe, but i won't clear the needs refresh flag. */
static inline void ocfs2_complete_lock_res_refresh(struct ocfs2_lock_res *lockres,
						   int status)
{
	mlog_entry_void();

	spin_lock(&lockres->l_lock);
	lockres_clear_flags(lockres, OCFS2_LOCK_REFRESHING);
	if (!status)
		lockres_clear_flags(lockres, OCFS2_LOCK_NEEDS_REFRESH);
	spin_unlock(&lockres->l_lock);

	wake_up(&lockres->l_event);

	mlog_exit_void();
}

/* may or may not return a bh if it went to disk. */
static int ocfs2_meta_lock_update(struct inode *inode,
				  struct buffer_head **bh)
{
	int status = 0;
	u32 trustable_clusters = 0;
	struct ocfs2_lock_res *lockres;
	ocfs2_dinode *fe;

	mlog_entry_void();

	spin_lock(&OCFS2_I(inode)->ip_lock);
	if (OCFS2_I(inode)->ip_flags & OCFS2_INODE_DELETED) {
		mlog(0, "Orphaned inode %"MLFu64" was deleted while we "
		     "were waiting on a lock. ip_flags = 0x%x\n",
		     OCFS2_I(inode)->ip_blkno, OCFS2_I(inode)->ip_flags);
		spin_unlock(&OCFS2_I(inode)->ip_lock);
		status = -ENOENT;
		goto bail;
	}
	spin_unlock(&OCFS2_I(inode)->ip_lock);

	lockres = &OCFS2_I(inode)->ip_meta_lockres;

	if (!ocfs2_should_refresh_lock_res(lockres))
		goto bail;

	/* This will discard any caching information we might have had
	 * for the inode metadata. */
	ocfs2_metadata_cache_purge(inode);

	if (ocfs2_lvb_is_trustable(lockres)) {
		/* yay, fastpath! */
		ocfs2_meta_lvb_get_trunc_clusters(lockres,
						  &trustable_clusters);
		ocfs2_refresh_inode_from_lvb(inode);
	} else {
		/* Boo, we have to go to disk. */
		/* read bh, cast, ocfs2_refresh_inode */
		status = ocfs2_read_block(OCFS2_SB(inode->i_sb),
					  OCFS2_I(inode)->ip_blkno, bh,
					  OCFS2_BH_CACHED, inode);
		if (status < 0) {
			mlog_errno(status);
			goto bail_refresh;
		}
		fe = (ocfs2_dinode *) (*bh)->b_data;

		/* This is a good chance to make sure we're not
		 * locking an invalid object.
		 *
		 * We bug on a stale inode here because we checked
		 * above whether it was wiped from disk. The wiping
		 * node provides a guarantee that we receive that
		 * message and can mark the inode before dropping any
		 * locks associated with it. */
		OCFS2_BUG_ON_INVALID_DINODE(fe);
		mlog_bug_on_msg(inode->i_generation !=
				le32_to_cpu(fe->i_generation),
				"Invalid dinode %"MLFu64" disk generation: %u "
				"inode->i_generation: %u\n",
				OCFS2_I(inode)->ip_blkno,
				le32_to_cpu(fe->i_generation),
				inode->i_generation);
		mlog_bug_on_msg(fe->i_dtime || !(fe->i_flags & OCFS2_VALID_FL),
				"Stale dinode %"MLFu64" dtime: %"MLFu64" "
				"flags: 0x%x\n", OCFS2_I(inode)->ip_blkno,
				fe->i_dtime, fe->i_flags);

		ocfs2_refresh_inode(inode, fe);
	}

	mlog(0, "inode %"MLFu64", I can only trust %u clusters\n",
	     OCFS2_I(inode)->ip_blkno, trustable_clusters);

	/* will do nothing for inode types that don't use the extent
	 * map (directories, bitmap files, etc) */
	ocfs2_extent_map_trunc(inode, trustable_clusters);

	ocfs2_set_local_seq_from_lvb(lockres);
	ocfs2_reset_meta_lvb_values(inode);

	status = 0;
bail_refresh:
	ocfs2_complete_lock_res_refresh(lockres, status);
bail:
	mlog_exit(status);
	return status;
}

static int ocfs2_assign_bh(struct inode *inode,
			   struct buffer_head **ret_bh,
			   struct buffer_head *passed_bh)
{
	int status;

	if (passed_bh) {
		/* Ok, the update went to disk for us, use the
		 * returned bh. */
		*ret_bh = passed_bh;
		get_bh(*ret_bh);

		return 0;
	}

	status = ocfs2_read_block(OCFS2_SB(inode->i_sb),
				  OCFS2_I(inode)->ip_blkno,
				  ret_bh,
				  OCFS2_BH_CACHED,
				  inode);
	if (status < 0)
		mlog_errno(status);

	return status;
}

/* 
 * returns < 0 error if the callback will never be called, otherwise
 * the result of the lock will be communicated via the callback.
 */
int ocfs2_meta_lock_full(struct inode *inode,
			 ocfs2_journal_handle *handle,
			 struct buffer_head **ret_bh,
			 int ex,
			 int flags,
			 ocfs2_lock_callback cb,
			 unsigned long cb_data)
{
	int status, level, dlm_flags, acquired;
	struct ocfs2_lock_res *lockres;
	ocfs2_super *osb = OCFS2_SB(inode->i_sb);
	struct buffer_head *local_bh = NULL;

	OCFS2_ASSERT(inode);

	mlog_entry_void();

	mlog(0, "inode %"MLFu64", take %s META lock\n",
	     OCFS2_I(inode)->ip_blkno,
	     ex ? "EXMODE" : "PRMODE");

	status = 0;
	acquired = 0;

	if (!(flags & OCFS2_META_LOCK_RECOVERY))
		status = wait_event_interruptible(osb->recovery_event,
						  ocfs2_node_map_is_empty(osb,
									  &osb->recovery_map));
	if (status < 0) {
		mlog_errno(status);
		return status;
	}

	lockres = &OCFS2_I(inode)->ip_meta_lockres;
	level = ex ? LKM_EXMODE : LKM_PRMODE;
	dlm_flags = 0;
	if (flags & OCFS2_META_LOCK_NOQUEUE)
		dlm_flags |= LKM_NOQUEUE;

	status = ocfs2_cluster_lock(osb, lockres, level, dlm_flags, cb,
				    cb_data);
	if (status < 0) {
		if (status != -EAGAIN)
			mlog_errno(status);
		goto bail;
	}

	/* Notify the error cleanup path to drop the cluster lock. */
	acquired = 1;

	/* We wait twice because a node may have died while we were in
	 * the lower dlm layers. The second time though, we've
	 * committed to owning this lock so we don't allow signals to
	 * abort the operation. */
	if (!(flags & OCFS2_META_LOCK_RECOVERY))
		wait_event(osb->recovery_event,
			   ocfs2_node_map_is_empty(osb,
						   &osb->recovery_map));

	/* This is fun. The caller may want a bh back, or it may
	 * not. ocfs2_meta_lock_update definitely wants one in, but
	 * may or may not read one, depending on what's in the
	 * LVB. The result of all of this is that we've *only* gone to
	 * disk if we have to, so the complexity is worthwhile. */
	status = ocfs2_meta_lock_update(inode, &local_bh);
	if (status < 0) {
		if (status != -ENOENT)
			mlog_errno(status);
		goto bail;
	}

	if (ret_bh) {
		status = ocfs2_assign_bh(inode, ret_bh, local_bh);
		if (status < 0) {
			mlog_errno(status);
			goto bail;
		}
	}

	if (handle) {
		status = ocfs2_handle_add_lock(handle, inode);
		if (status < 0)
			mlog_errno(status);
	}

bail:
	if (status < 0) {
		if (ret_bh && (*ret_bh)) {
			brelse(*ret_bh);
			*ret_bh = NULL;
		}
		if (acquired)
			ocfs2_meta_unlock(inode, ex);
	}

	if (local_bh)
		brelse(local_bh);

	mlog_exit(status);
	return status;
}

void ocfs2_meta_unlock(struct inode *inode,
		       int ex)
{
	int level = ex ? LKM_EXMODE : LKM_PRMODE;
	struct ocfs2_lock_res *lockres = &OCFS2_I(inode)->ip_meta_lockres;

	mlog_entry_void();

	mlog(0, "inode %"MLFu64" drop %s META lock\n",
	     OCFS2_I(inode)->ip_blkno,
	     ex ? "EXMODE" : "PRMODE");

	ocfs2_cluster_unlock(OCFS2_SB(inode->i_sb), lockres, level);

	mlog_exit_void();
}

int ocfs2_super_lock(ocfs2_super *osb,
		     int ex)
{
	int status;
	int level = ex ? LKM_EXMODE : LKM_PRMODE;
	struct ocfs2_lock_res *lockres = &osb->osb_super_lockres;
	struct buffer_head *bh;
	ocfs2_slot_info *si = osb->slot_info;

	mlog_entry_void();

	status = ocfs2_cluster_lock(osb, lockres, level, 0, NULL, 0);
	if (status < 0) {
		mlog_errno(status);
		goto bail;
	}

	/* The super block lock path is really in the best position to
	 * know when resources covered by the lock need to be
	 * refreshed, so we do it here. Of course, making sense of
	 * everything is up to the caller :) */
	status = ocfs2_should_refresh_lock_res(lockres);
	if (status < 0) {
		mlog_errno(status);
		goto bail;
	}
	if (status) {
		bh = si->si_bh;
		status = ocfs2_read_block(osb, bh->b_blocknr, &bh, 0,
					  si->si_inode);
		if (status == 0)
			ocfs2_update_slot_info(si);

		ocfs2_complete_lock_res_refresh(lockres, status);

		if (status < 0)
			mlog_errno(status);
	}
bail:
	mlog_exit(status);
	return status;
}

void ocfs2_super_unlock(ocfs2_super *osb,
			int ex)
{
	int level = ex ? LKM_EXMODE : LKM_PRMODE;
	struct ocfs2_lock_res *lockres = &osb->osb_super_lockres;

	ocfs2_cluster_unlock(osb, lockres, level);
}

int ocfs2_rename_lock(ocfs2_super *osb)
{
	int status;
	struct ocfs2_lock_res *lockres = &osb->osb_rename_lockres;

	status = ocfs2_cluster_lock(osb, lockres, LKM_EXMODE, 0, NULL, 0);
	if (status < 0)
		mlog_errno(status);

	return status;
}

void ocfs2_rename_unlock(ocfs2_super *osb)
{
	struct ocfs2_lock_res *lockres = &osb->osb_rename_lockres;

	ocfs2_cluster_unlock(osb, lockres, LKM_EXMODE);
}

int ocfs2_dlm_init(ocfs2_super *osb)
{
	int status;
	u32 dlm_key;
	dlm_ctxt *dlm = NULL;

	mlog_entry_void();

	/* launch vote thread */
	osb->vote_task = kthread_run(ocfs2_vote_thread, osb, "ocfs2vote-%d",
				     osb->osb_id);
	if (IS_ERR(osb->vote_task)) {
		status = PTR_ERR(osb->vote_task);
		osb->vote_task = NULL;
		mlog_errno(status);
		goto bail;
	}

	/* used by the dlm code to make message headers unique, each
	 * node in this domain must agree on this. */
	dlm_key = crc32(0, osb->uuid_str, strlen(osb->uuid_str));

	/* for now, uuid == domain */
	dlm = dlm_register_domain(osb->uuid_str, dlm_key);
	if (!dlm) {
		/* This is a best guess on return value... */
		status = -ENOMEM;
		mlog_errno(status);
		goto bail;
	}

	ocfs2_super_lock_res_init(&osb->osb_super_lockres, osb);
	ocfs2_rename_lock_res_init(&osb->osb_rename_lockres, osb);

	dlm_register_eviction_cb(dlm, &osb->osb_eviction_cb);

	osb->dlm = dlm;

	status = 0;
bail:

	mlog_exit(status);
	return status;
}

void ocfs2_dlm_shutdown(ocfs2_super *osb)
{
	mlog_entry_void();

	dlm_unregister_eviction_cb(&osb->osb_eviction_cb);

	ocfs2_drop_osb_locks(osb);

	if (osb->vote_task) {
		kthread_stop(osb->vote_task);
		osb->vote_task = NULL;
	}

	ocfs2_lock_res_free(&osb->osb_super_lockres);
	ocfs2_lock_res_free(&osb->osb_rename_lockres);

	dlm_unregister_domain(osb->dlm);
	osb->dlm = NULL;

	mlog_exit_void();
}

static void ocfs2_unlock_ast_func(void *opaque, dlm_status status)
{
	struct ocfs2_lock_res *lockres = opaque;

	mlog_entry_void();

	mlog(0, "UNLOCK AST called on lock %s, action = %d\n", lockres->l_name,
	     lockres->l_unlock_action);

	if (status != DLM_NORMAL)
		mlog(ML_ERROR, "Dlm returns status %d for lock %s\n",
		     status, lockres->l_name);

	spin_lock(&lockres->l_lock);
	switch(lockres->l_unlock_action) {
	case OCFS2_UNLOCK_CANCEL_CONVERT:
		lockres->l_action = OCFS2_AST_INVALID;
		break;
	case OCFS2_UNLOCK_DROP_LOCK:
		lockres->l_level = LKM_IVMODE;
		break;
	default:
		BUG();
	}
	lockres->l_unlock_action = OCFS2_UNLOCK_INVALID;
	lockres_clear_flags(lockres, OCFS2_LOCK_BUSY);
	spin_unlock(&lockres->l_lock);

	wake_up(&lockres->l_event);

	mlog_exit_void();
}


/* BEWARE: called with lockres lock, and always drops it. Caller
 * should not be calling us with a busy lock... */
static int __ocfs2_drop_lock(ocfs2_super *osb,
			     struct ocfs2_lock_res *lockres)
{
	int ret = 0;
	dlm_status status;

	if (lockres->l_flags & OCFS2_LOCK_BUSY)
		mlog(ML_ERROR, "destroying busy lock: \"%s\"\n",
		     lockres->l_name);
	if (lockres->l_flags & OCFS2_LOCK_BLOCKED)
		mlog(0, "destroying blocked lock: \"%s\"\n", lockres->l_name);

	if (!(lockres->l_flags & OCFS2_LOCK_ATTACHED)) {
		spin_unlock(&lockres->l_lock);
		goto bail;
	}

	lockres_clear_flags(lockres, OCFS2_LOCK_ATTACHED);

	/* make sure we never get here while waiting for an ast to
	 * fire. */
	OCFS2_ASSERT(lockres->l_action == OCFS2_AST_INVALID);

	/* is this necessary? */
	lockres_or_flags(lockres, OCFS2_LOCK_BUSY);
	lockres->l_unlock_action = OCFS2_UNLOCK_DROP_LOCK;
	spin_unlock(&lockres->l_lock);

	mlog(0, "lock %s\n", lockres->l_name);

	status = dlmunlock(osb->dlm,
			   &lockres->l_lksb,
			   LKM_VALBLK,
			   lockres->l_ops->unlock_ast,
			   lockres);
	if (status != DLM_NORMAL) {
		ocfs2_log_dlm_error("dlmunlock", status, lockres);
		mlog(ML_ERROR, "lockres flags: %lu\n", lockres->l_flags);
		dlm_print_one_lock(lockres->l_lksb.lockid);
		BUG();
	}
	mlog(0, "lock %s, successfull return from dlmunlock\n",
	     lockres->l_name);

	ocfs2_wait_on_busy_lock(lockres);
bail:
	mlog_exit(ret);
	return ret;
}

typedef void (ocfs2_pre_drop_cb_t)(struct ocfs2_lock_res *, void *);

struct drop_lock_cb {
	ocfs2_pre_drop_cb_t	*drop_func;
	void			*drop_data;
};

static int ocfs2_drop_lock(ocfs2_super *osb,
			   struct ocfs2_lock_res *lockres,
			   struct drop_lock_cb *dcb)
{
	/* We didn't get anywhere near actually using this lockres. */
	if (!(lockres->l_flags & OCFS2_LOCK_INITIALIZED))
		return 0;

	spin_lock(&lockres->l_lock);

	while (lockres->l_flags & OCFS2_LOCK_BUSY) {
		mlog(0, "waiting on busy lock \"%s\": flags = %lx, action = "
		     "%u, unlock_action = %u\n",
		     lockres->l_name, lockres->l_flags, lockres->l_action,
		     lockres->l_unlock_action);

		spin_unlock(&lockres->l_lock);

		/* XXX: Today we just wait on any busy
		 * locks... Perhaps we need to cancel converts in the
		 * future? */
		ocfs2_wait_on_busy_lock(lockres);

		spin_lock(&lockres->l_lock);
	}

	if (dcb)
		dcb->drop_func(lockres, dcb->drop_data);

	/* This will drop the spinlock for us. Dur de dur, at least we
	 * keep the ugliness in one place :) */
	return  __ocfs2_drop_lock(osb, lockres);
}

static void ocfs2_drop_osb_locks(ocfs2_super *osb)
{
	int status;

	mlog_entry_void();

	status = ocfs2_drop_lock(osb, &osb->osb_super_lockres, NULL);
	if (status < 0)
		mlog_errno(status);

	status = ocfs2_drop_lock(osb, &osb->osb_rename_lockres, NULL);
	if (status < 0)
		mlog_errno(status);

	mlog_exit(status);
}

static void ocfs2_meta_pre_drop(struct ocfs2_lock_res *lockres, void *data)
{
	struct inode *inode = data;

	/* the metadata lock requires a bit more work as we have an
	 * LVB to worry about. */
	if (lockres->l_flags & OCFS2_LOCK_ATTACHED) {
		if (lockres->l_level == LKM_EXMODE)
			__ocfs2_stuff_meta_lvb(inode);
		/* fake an NLMODE downconvert for the lvb code. */
		__ocfs2_lvb_on_downconvert(lockres, LKM_NLMODE);
	}
}

int ocfs2_drop_inode_locks(struct inode *inode)
{
	int status, err;
	struct drop_lock_cb meta_dcb = { ocfs2_meta_pre_drop, inode, };

	mlog_entry_void();

	err = ocfs2_drop_lock(OCFS2_SB(inode->i_sb),
			      &OCFS2_I(inode)->ip_data_lockres,
			      NULL);
	if (err < 0)
		mlog_errno(err);

	status = err;

	err = ocfs2_drop_lock(OCFS2_SB(inode->i_sb),
			      &OCFS2_I(inode)->ip_meta_lockres,
			      &meta_dcb);
	if (err < 0)
		mlog_errno(err);
	if (err < 0 && !status)
		status = err;

	mlog_exit(status);
	return status;
}

/* called with the spinlock held, and WILL drop it. */
static int __ocfs2_downconvert_lock(ocfs2_super *osb,
				    struct ocfs2_lock_res *lockres,
				    int new_level,
				    int lvb)
{
	int ret, flags = LKM_CONVERT;
	dlm_status status;

	mlog_entry_void();

	OCFS2_ASSERT(lockres->l_blocking > LKM_NLMODE);

	if (lockres->l_level <= new_level) {
		mlog(ML_ERROR, "lockres->l_level (%u) <= new_level (%u)\n",
		     lockres->l_level, new_level);
		BUG();
	}

	mlog(0, "lock %s, new_level = %d, l_blocking = %d, lvb = %d\n",
	     lockres->l_name, new_level, lockres->l_blocking, lvb);

	lockres->l_action = OCFS2_AST_DOWNCONVERT;
	lockres->l_requested = new_level;
	lockres_or_flags(lockres, OCFS2_LOCK_BUSY);
	spin_unlock(&lockres->l_lock);

	if (lvb)
		flags |= LKM_VALBLK;

	status = dlmlock(osb->dlm,
			 new_level,
			 &lockres->l_lksb,
			 flags,
			 lockres->l_name,
			 lockres->l_ops->ast,
			 lockres,
			 lockres->l_ops->bast);
	if (status != DLM_NORMAL) {
		ocfs2_log_dlm_error("dlmlock", status, lockres);
		ret = -EINVAL;
		ocfs2_recover_from_dlm_error(lockres, 1);
		goto bail;
	}

	ret = 0;
bail:
	mlog_exit(ret);
	return ret;
}

/* called with the spinlock held, and WILL drop it. */
static int __ocfs2_cancel_convert(ocfs2_super *osb,
				  struct ocfs2_lock_res *lockres)
{
	int ret;
	dlm_status status;

	mlog_entry_void();

	mlog(0, "lock %s\n", lockres->l_name);

	/* were we in a convert when we got the bast fire? */
	OCFS2_ASSERT(lockres->l_action == OCFS2_AST_CONVERT ||
		    lockres->l_action == OCFS2_AST_DOWNCONVERT);
	/* set things up for the unlockast to know to just
	 * clear out the ast_action and unset busy, etc. */
	lockres->l_unlock_action = OCFS2_UNLOCK_CANCEL_CONVERT;
	spin_unlock(&lockres->l_lock);

	ret = 0;
	status = dlmunlock(osb->dlm,
			   &lockres->l_lksb,
			   LKM_CANCEL,
			   lockres->l_ops->unlock_ast,
			   lockres);
	if (status == DLM_CANCELGRANT) {
		/* If we got this, then the ast was fired
		 * before we could cancel. We cleanup our
		 * state, and restart the function. */
		spin_lock(&lockres->l_lock);
		/* Can we race lock destruction here? Probably not but
		 * lets double check just in case... */
		if (lockres->l_unlock_action == OCFS2_UNLOCK_CANCEL_CONVERT)
			lockres->l_unlock_action = OCFS2_UNLOCK_INVALID;
		spin_unlock(&lockres->l_lock);
	} else if (status != DLM_NORMAL) {
		ocfs2_log_dlm_error("dlmunlock", status, lockres);
		ret = -EINVAL;
		ocfs2_recover_from_dlm_error(lockres, 0);
	}

	mlog(0, "lock %s return from dlmunlock\n", lockres->l_name);

	mlog_exit(ret);
	return ret;
}

static inline int ocfs2_can_downconvert_meta_lock(struct inode *inode,
						  struct ocfs2_lock_res *lockres,
						  int new_level)
{
	int ret;

	mlog_entry_void();

	OCFS2_ASSERT(new_level == LKM_NLMODE || new_level == LKM_PRMODE);

	if (lockres->l_flags & OCFS2_LOCK_REFRESHING) {
		ret = 0;
		mlog(0, "lockres %s currently being refreshed -- backing "
		     "off!\n", lockres->l_name);
	} else if (new_level == LKM_PRMODE)
		ret = !lockres->l_ex_holders && 
			ocfs2_inode_fully_checkpointed(inode);
	else /* Must be NLMODE we're converting to. */
		ret = !lockres->l_ro_holders && !lockres->l_ex_holders &&
			ocfs2_inode_fully_checkpointed(inode);

	mlog_exit(ret);
	return ret;
}

static int ocfs2_do_unblock_meta(struct inode *inode,
				 int *requeue)
{
	int new_level;
	int set_lvb = 0;
	int ret = 0;
	struct ocfs2_lock_res *lockres;

	mlog_entry_void();

       	lockres = &OCFS2_I(inode)->ip_meta_lockres;

	spin_lock(&lockres->l_lock);

	OCFS2_ASSERT(lockres->l_flags & OCFS2_LOCK_BLOCKED);

	mlog(0, "l_level=%d, l_blocking=%d\n", lockres->l_level,
	     lockres->l_blocking);

	OCFS2_ASSERT(lockres->l_level == LKM_EXMODE || 
		    lockres->l_level == LKM_PRMODE);

	if (lockres->l_flags & OCFS2_LOCK_BUSY) {
		*requeue = 1;
		if (lockres->l_unlock_action == OCFS2_UNLOCK_CANCEL_CONVERT) {
			/* If we're already trying to cancel a lock conversion
			 * then just drop the spinlock and requeue ourselves
			 * to check again later. */
			spin_unlock(&lockres->l_lock);
			ret = 0;
			goto leave;
		}

		ret = __ocfs2_cancel_convert(OCFS2_SB(inode->i_sb),
					      lockres);
		goto leave;
	}

	new_level = ocfs2_highest_compat_lock_level(lockres->l_blocking);

	mlog(0, "l_level=%d, l_blocking=%d, new_level=%d\n",
	     lockres->l_level, lockres->l_blocking, new_level);

	if (ocfs2_can_downconvert_meta_lock(inode, lockres, new_level)) {
		if (lockres->l_level == LKM_EXMODE)
			set_lvb = 1;

		/* If the lock hasn't been refreshed yet (rare), then our
		 * memory inode values are old and we skip stuffing
		 * the lvb. Additionally we cannot incrememnt the lvb
		 * sequence numbers */
		if (!(lockres->l_flags & OCFS2_LOCK_NEEDS_REFRESH)) {
			if (set_lvb)
				__ocfs2_stuff_meta_lvb(inode);
			__ocfs2_lvb_on_downconvert(lockres, new_level);
		} else
			mlog(0, "lockres %s: downconverting stale lock!\n",
			     lockres->l_name);

		mlog(0, "calling __ocfs2_downconvert_lock with "
		     "l_level=%d, l_blocking=%d, new_level=%d\n",
		     lockres->l_level, lockres->l_blocking,
		     new_level);
		ret = __ocfs2_downconvert_lock(OCFS2_SB(inode->i_sb),
					       lockres, new_level,
					       set_lvb);
		goto leave;
	}
	if (!ocfs2_inode_fully_checkpointed(inode))
		ocfs2_start_checkpoint(OCFS2_SB(inode->i_sb));

	*requeue = 1;
	spin_unlock(&lockres->l_lock);
	ret = 0;
leave:
	mlog_exit(ret);
	return ret;
}

static int ocfs2_generic_unblock_lock(ocfs2_super *osb,
				      struct ocfs2_lock_res *lockres,
				      int *requeue,
				      ocfs2_convert_worker_t *worker)
{
	int blocking;
	int new_level;
	int ret = 0;

	mlog_entry_void();

	spin_lock(&lockres->l_lock);

	OCFS2_ASSERT(lockres->l_flags & OCFS2_LOCK_BLOCKED);

recheck:
	if (lockres->l_flags & OCFS2_LOCK_BUSY) {
		*requeue = 1;
		if (lockres->l_unlock_action == OCFS2_UNLOCK_CANCEL_CONVERT) {
			/* If we're already trying to cancel a lock conversion
			 * then just drop the spinlock and requeue ourselves
			 * to check again later. */
			spin_unlock(&lockres->l_lock);
			ret = 0;
			goto leave;
		}

		ret = __ocfs2_cancel_convert(osb, lockres);
		goto leave;
	}

	/* if we're blocking an exclusive and we have *any* holders,
	 * then requeue. */
	if ((lockres->l_blocking == LKM_EXMODE) 
	    && (lockres->l_ex_holders || lockres->l_ro_holders)) {
		spin_unlock(&lockres->l_lock);
		*requeue = 1;
		ret = 0;
		goto leave;
	}

	/* If it's a PR we're blocking, then only
	 * requeue if we've got any EX holders */
	if (lockres->l_blocking == LKM_PRMODE &&
	    lockres->l_ex_holders) {
		spin_unlock(&lockres->l_lock);
		*requeue = 1;
		ret = 0;
		goto leave;
	}

	/* If we get here, then we know that there are no more
	 * incompatible holders (and anyone asking for an incompatible
	 * lock is blocked). We can now downconvert the lock */
	if (!worker)
		goto downconvert;

	/* Some lockres types want to do a bit of work before
	 * downconverting a lock. Allow that here. The worker function
	 * may sleep, so we save off a copy of what we're blocking as
	 * it may change while we're not holding the spin lock. */
	blocking = lockres->l_blocking;
	spin_unlock(&lockres->l_lock);

	worker(lockres, blocking);

	spin_lock(&lockres->l_lock);
	if (blocking != lockres->l_blocking) {
		/* If this changed underneath us, then we can't drop
		 * it just yet. */
		goto recheck;
	}

downconvert:
	*requeue = 0;
	new_level = ocfs2_highest_compat_lock_level(lockres->l_blocking);

	ret = __ocfs2_downconvert_lock(osb, lockres, new_level, 0);
leave:
	mlog_exit(ret);
	return ret;
}

static void ocfs2_data_convert_worker(struct ocfs2_lock_res *lockres,
				      int blocking)
{
	struct inode *inode;
	struct address_space *mapping;

	mlog_entry_void();

       	inode = ocfs2_lock_res_inode(lockres);
	mapping = inode->i_mapping;

	if (filemap_fdatawrite(mapping)) {
		mlog(ML_ERROR, "Could not sync inode %"MLFu64" for downconvert!",
		     OCFS2_I(inode)->ip_blkno);
	}
	sync_mapping_buffers(mapping);
	if (blocking == LKM_EXMODE) {
		truncate_inode_pages(mapping, 0);
		unmap_mapping_range(mapping, 0, 0, 0);
	} else {
		/* We only need to wait on the I/O if we're not also
		 * truncating pages because truncate_inode_pages waits
		 * for us above. We don't truncate pages if we're
		 * blocking anything < EXMODE because we want to keep
		 * them around in that case. */
		filemap_fdatawait(mapping);
	}

	mlog_exit_void();
}

int ocfs2_unblock_data(struct ocfs2_lock_res *lockres,
		       int *requeue)
{
	int status;
	struct inode *inode;
	ocfs2_super *osb;

	mlog_entry_void();

	inode = ocfs2_lock_res_inode(lockres);
	osb = OCFS2_SB(inode->i_sb);

	mlog(0, "unblock inode %"MLFu64"\n", OCFS2_I(inode)->ip_blkno);

	status = ocfs2_generic_unblock_lock(osb,
					    lockres,
					    requeue,
					    ocfs2_data_convert_worker);
	if (status < 0)
		mlog_errno(status);

	mlog(0, "inode %"MLFu64", requeue = %d\n",
	     OCFS2_I(inode)->ip_blkno, *requeue);

	/* because of inode ref counting, we never want to propagate
	 * up requeue requests for inode locks. Instead we do it
	 * ourselves here, and lose the extra ref we got from queueing
	 * when we came in. */
	if (*requeue)
		ocfs2_schedule_blocked_inode_lock(inode, lockres);

	iput(inode);
	*requeue = 0;

	mlog_exit(status);
	return status;
}

int ocfs2_unblock_meta(struct ocfs2_lock_res *lockres,
		       int *requeue)
{
	int status;
	struct inode *inode;

	mlog_entry_void();

       	inode = ocfs2_lock_res_inode(lockres);

	mlog(0, "unblock inode %"MLFu64"\n", OCFS2_I(inode)->ip_blkno);

	status = ocfs2_do_unblock_meta(inode, requeue);
	if (status < 0)
		mlog_errno(status);

	mlog(0, "inode %"MLFu64", requeue = %d\n",
	     OCFS2_I(inode)->ip_blkno, *requeue);

	/* if you're confused by this, see the comment in
	 * ocfs2_unblock_data */
	if (*requeue)
		ocfs2_schedule_blocked_inode_lock(inode, lockres);

	iput(inode);
	*requeue = 0;

	mlog_exit(status);
	return status;
}

/* Generic unblock function for any lockres whose private data is an
 * ocfs2_super pointer. */
static int ocfs2_unblock_osb_lock(struct ocfs2_lock_res *lockres,
				  int *requeue)
{
	int status;
	ocfs2_super *osb;

	mlog_entry_void();

	mlog(0, "Unblock lockres %s\n", lockres->l_name);

	osb = ocfs2_lock_res_super(lockres);

	status = ocfs2_generic_unblock_lock(osb,
					    lockres,
					    requeue,
					    NULL);
	if (status < 0)
		mlog_errno(status);

	mlog_exit(status);
	return status;
}

void ocfs2_process_blocked_lock(ocfs2_super *osb,
				struct ocfs2_lock_res *lockres)
{
	int status;
	int requeue = 0;

	mlog_entry_void();

	OCFS2_ASSERT(lockres);
	OCFS2_ASSERT(lockres->l_ops);
	OCFS2_ASSERT(lockres->l_ops->unblock);

	mlog(0, "lockres %s blocked.\n", lockres->l_name);

	status = lockres->l_ops->unblock(lockres, &requeue);
	if (status < 0)
		mlog_errno(status);

	if (requeue)
		ocfs2_schedule_blocked_lock(osb, lockres);

	mlog(0, "lockres %s, requeue = %s.\n", lockres->l_name,
	     requeue ? "yes" : "no");

	mlog_exit_void();
}

static void ocfs2_schedule_blocked_lock(ocfs2_super *osb,
					struct ocfs2_lock_res *lockres)
{
	mlog_entry_void();

	spin_lock(&osb->vote_task_lock);
	if (list_empty(&lockres->l_blocked_list)) {
		list_add_tail(&lockres->l_blocked_list,
			      &osb->blocked_lock_list);
		osb->blocked_lock_count++;
	}
	spin_unlock(&osb->vote_task_lock);

	mlog_exit_void();
}

/* needed for inodes as we have to take a reference on them.. */
static void ocfs2_schedule_blocked_inode_lock(struct inode *inode,
					      struct ocfs2_lock_res *lockres)
{
	mlog_entry_void();

	if (!igrab(inode)) {
		mlog(0, "Inode %"MLFu64" asked to be scheduled during "
		     "clear_inode!\n", OCFS2_I(inode)->ip_blkno);
		return;
	}

	ocfs2_schedule_blocked_lock(OCFS2_SB(inode->i_sb), lockres);

	mlog_exit_void();
}

void ocfs2_meta_lvb_set_trunc_clusters(struct inode *inode,
				       unsigned int trunc_clusters)
{
	struct ocfs2_lock_res *lockres = &OCFS2_I(inode)->ip_meta_lockres;
	struct ocfs2_meta_lvb *lvb;

	spin_lock(&lockres->l_lock);
	OCFS2_ASSERT(lockres->l_level == LKM_EXMODE);

	lvb = (struct ocfs2_meta_lvb *) lockres->l_lksb.lvb;
	if (be32_to_cpu(lvb->lvb_trunc_clusters) > trunc_clusters)
		lvb->lvb_trunc_clusters = cpu_to_be32(trunc_clusters);
	spin_unlock(&lockres->l_lock);
}
