/*
	kmod, the new module loader (replaces kerneld)
	Kirk Petersen

	Reorganized not to be a daemon by Adam Richter, with guidance
	from Greg Zornetzer.

	Modified to avoid chroot and file sharing problems.
	Mikael Pettersson

	Limit the concurrent number of kmod modprobes to catch loops from
	"modprobe needs a service that is in a module".
	Keith Owens <kaos@ocs.com.au> December 1999

	Unblock all signals when we exec a usermode process.
	Shuu Yamaguchi <shuu@wondernetworkresources.com> December 2000

	call_usermodehelper wait flag, and remove exec_usermodehelper.
	Rusty Russell <rusty@rustcorp.com.au>  Jan 2003

	resource management for call_usermodehelper
	Hannes Reinecke <hare@suse.de> Jul 2004
*/
#define __KERNEL_SYSCALLS__

#include <linux/config.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/syscalls.h>
#include <linux/unistd.h>
#include <linux/kmod.h>
#include <linux/smp_lock.h>
#include <linux/slab.h>
#include <linux/namespace.h>
#include <linux/completion.h>
#include <linux/file.h>
#include <linux/workqueue.h>
#include <linux/security.h>
#include <linux/mount.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <asm/uaccess.h>

#undef DEBUG_KHELPER

extern int max_threads;

#ifdef CONFIG_KMOD
static struct workqueue_struct *khelper_wq;
int khelper_max = 64;
static struct semaphore khelper_sem;

/*
	modprobe_path is set via /proc/sys.
*/
char modprobe_path[256] = "/sbin/modprobe";

/**
 * request_module - try to load a kernel module
 * @fmt:     printf style format string for the name of the module
 * @varargs: arguements as specified in the format string
 *
 * Load a module using the user mode module loader. The function returns
 * zero on success or a negative errno code on failure. Note that a
 * successful module load does not mean the module did not then unload
 * and exit on an error of its own. Callers must check that the service
 * they requested is now available not blindly invoke it.
 *
 * If module auto-loading support is disabled then this function
 * becomes a no-operation.
 */
int request_module(const char *fmt, ...)
{
	va_list args;
	char module_name[MODULE_NAME_LEN];
	int ret;
	char *argv[] = { modprobe_path, "-q", "--", module_name, NULL };
	static char *envp[] = { "HOME=/",
				"TERM=linux",
				"PATH=/sbin:/usr/sbin:/bin:/usr/bin",
				NULL };

	va_start(args, fmt);
	ret = vsnprintf(module_name, MODULE_NAME_LEN, fmt, args);
	va_end(args);
	if (ret >= MODULE_NAME_LEN)
		return -ENAMETOOLONG;

	/* If modprobe needs a service that is in a module, we get a recursive
	 * loop.  Limit the number of running kmod threads to max_threads/2 or
	 * MAX_KMOD_CONCURRENT, whichever is the smaller.  A cleaner method
	 * would be to run the parents of this process, counting how many times
	 * kmod was invoked.  That would mean accessing the internals of the
	 * process tables to get the command line, proc_pid_cmdline is static
	 * and it is not worth changing the proc code just to handle this case. 
	 * KAO.
	 *
	 * "trace the ppid" is simple, but will fail if someone's
	 * parent exits.  I think this is as good as it gets. --RR
	 *
	 * Resource checking is now implemented in 
	 * call_usermodehelper --hare
	 */
	ret = call_usermodehelper(modprobe_path, argv, envp, 1);
	return ret;
}
EXPORT_SYMBOL(request_module);
#endif /* CONFIG_KMOD */

#ifdef CONFIG_HOTPLUG
/*
	hotplug path is set via /proc/sys
	invoked by hotplug-aware bus drivers,
	with call_usermodehelper

	argv [0] = hotplug_path;
	argv [1] = "usb", "scsi", "pci", "network", etc;
	... plus optional type-specific parameters
	argv [n] = 0;

	envp [*] = HOME, PATH; optional type-specific parameters

	a hotplug bus should invoke this for device add/remove
	events.  the command is expected to load drivers when
	necessary, and may perform additional system setup.
*/
char hotplug_path[256] = "/sbin/hotplug";

EXPORT_SYMBOL(hotplug_path);

#endif /* CONFIG_HOTPLUG */

struct subprocess_info {
	struct completion *complete;
	char *path;
	char **argv;
	char **envp;
	int wait;
	int retval;
};

/* 
 * Copy process args
 */
#define KHELPER_ARGSIZE 4096
#define KHELPER_ALIGNMENT 8

static struct subprocess_info *
khelper_copy_info(struct subprocess_info *orig)
{
	struct subprocess_info *new;
	int i, l, n;
	char *p;

	/* We always allocate an entire page */
	new = kmalloc(KHELPER_ARGSIZE, GFP_KERNEL);
	if (!new)
		return NULL;
	memset(new, 0, KHELPER_ARGSIZE);
	p = (char *)new;
	p += ALIGN(sizeof(struct subprocess_info),KHELPER_ALIGNMENT);

	n = strlen(orig->path) + 1;
	strcpy(p, orig->path);
	new->path = p;
	p += ALIGN(n,KHELPER_ALIGNMENT);

	/* Count the number of arguments */
	i = 0;
	while (orig->argv[i]) i++;

	new->argv = (char **)p;
	/* We allow up to 7 arguments */
	l = i < 7 ? i : 7;
	p += sizeof(char *) * (l + 1);

	i = 0;
	while (orig->argv[i]) {
		n = strlen(orig->argv[i]) + 1;
		if ( i >= l || (p - new->path) + n >= KHELPER_ARGSIZE )
			break;
		new->argv[i] = p;
		strcpy(p, orig->argv[i]);
		p += ALIGN(n,KHELPER_ALIGNMENT);
		i++;
	}
	new->argv[i] = NULL;

	/* Count the number of environment variables */
	i = 0;
	while (orig->envp[i]) i++;

	new->envp = (char **)p;
	/* And up to 31 environment variables */
	l = i < 31 ? i : 31;
	p += sizeof(char *) * (l + 1);

	i = 0;
	while (orig->envp[i]) {
		n = strlen(orig->envp[i]) + 1;
		if ( i >= l || (p - new->path) + n >= KHELPER_ARGSIZE )
			break;

		new->envp[i] = p;
		strcpy(p, orig->envp[i]);
		p += ALIGN(n,KHELPER_ALIGNMENT);
		i++;
	}
	new->envp[i] = NULL;

	return new;
}

/*
 * This is the task which runs the usermode application
 */
static int ____call_usermodehelper(void *data)
{
	struct subprocess_info *sub_info = data;
	int retval;
	cpumask_t mask = CPU_MASK_ALL;

	/* Unblock all signals. */
	flush_signals(current);
	spin_lock_irq(&current->sighand->siglock);
	flush_signal_handlers(current, 1);
	sigemptyset(&current->blocked);
	recalc_sigpending();
	spin_unlock_irq(&current->sighand->siglock);

	/* We can run anywhere, unlike our parent keventd(). */
	set_cpus_allowed(current, mask);

	retval = -EPERM;
	if (current->fs->root)
		retval = execve(sub_info->path, sub_info->argv,sub_info->envp);

	/* Exec failed? */
	sub_info->retval = retval;
	do_exit(0);
}

/* Keventd can't block, but this (a child) can. */
static int wait_for_helper(void *data)
{
	struct subprocess_info *sub_info = data;
	pid_t pid;
	struct k_sigaction sa;
	int flags = SIGCHLD, retval, wait = sub_info->wait;
#ifdef DEBUG_KHELPER
	char *ev_descr;
#endif

#ifdef DEBUG_KHELPER
	if (wait < 1) {
		ev_descr = sub_info->envp[3] + 7;
	} else {
		ev_descr = sub_info->argv[3];
	}
#endif

	/* Do the real action */
#ifdef DEBUG_KHELPER
	printk(KERN_DEBUG "khelper: exec event %s (current %d, max %d, wait %d)\n",
	       ev_descr, khelper_max - atomic_read(&khelper_sem.count), 
	       khelper_max, wait);
#endif

	/* Install a handler: if SIGCLD isn't handled sys_wait4 won't
	 * populate the status, but will return -ECHILD. */
	sa.sa.sa_handler = SIG_IGN;
	sa.sa.sa_flags = 0;
	siginitset(&sa.sa.sa_mask, sigmask(SIGCHLD));
	do_sigaction(SIGCHLD, &sa, (struct k_sigaction *)0);
	allow_signal(SIGCHLD);

	if (wait < 1) {
		/* CLONE_VFORK: wait until the usermode helper has execve'd
		 * successfully We need the data structures to stay around
		 * until that is done.  */
		flags |= CLONE_VFORK;
	}

	pid = kernel_thread(____call_usermodehelper, sub_info, flags);
	if (pid >= 0) {
		/* 
		 * usermodehelper started successfully
		 * We always block for the child to exit as we want to
		 * keep track of the number of currently running processes.
		 */

		sys_wait4(pid, (int __user *) &retval, 0, NULL);
	} else {
#ifdef DEBUG_KHELPER
		printk(KERN_INFO "khelper: exec event %s failed (%d)\n",
		       ev_descr, pid);
#endif
		retval = pid;
	}

	if (wait > 0) {
		/* 
		 * For synchronous events we can return the exit
		 * status of the child.
		 */
		sub_info->retval = retval;
		complete(sub_info->complete);
	} else {
		kfree(sub_info);
	}

	up(&khelper_sem);

	return 0;
}

/* This is run by khelper thread for synchronous events */
static void __call_usermodehelper(void *data)
{
	struct subprocess_info *sub_info = data;
	pid_t pid;

	/* Wait for the semaphore to become available */
	down(&khelper_sem);

	pid = kernel_thread(wait_for_helper, sub_info,
			    CLONE_FS | CLONE_FILES | SIGCHLD);

	if (pid < 0) {
		sub_info->retval = pid;
		complete(sub_info->complete);
		up(&khelper_sem);
	}
}

/* This is run by khelper thread for asynchronous events */
static void __call_usermodehelper_async(void *data)
{
	struct subprocess_info *sub_info = data, *stored_info;
	pid_t pid;
#ifdef DEBUG_KHELPER
	char *ev_descr;
#endif

	stored_info = khelper_copy_info(sub_info);
	if (!stored_info) {
		sub_info->retval = -ENOMEM;
		complete(sub_info->complete);
		return;
	}
	stored_info->wait = sub_info->wait;
	stored_info->complete = NULL;

#ifdef DEBUG_KHELPER
	if (stored_info->wait < 1) {
		ev_descr = stored_info->envp[3] + 7;
	} else {
		ev_descr = stored_info->argv[3];
	}
#endif

	/* Queue this event */
	sub_info->retval = 0;
	complete(sub_info->complete);

	if (down_trylock(&khelper_sem)) {
		/* 
		 * We have exceeded the maximum number of
		 * concurrent kmod invocations. Delay this process
		 * until enough resources are available again.
		 */
#ifdef DEBUG_KHELPER
		printk(KERN_DEBUG "khelper: delay event %s (current %d, max %d)\n",
		       ev_descr, khelper_max - atomic_read(&khelper_sem.count), 
		       khelper_max);
#endif
		/* Wait for the semaphore to become available */
		down(&khelper_sem);
	}

	pid = kernel_thread(wait_for_helper, stored_info,
			    CLONE_FS | CLONE_FILES | SIGCHLD );

	if (pid < 0) {
		kfree(stored_info);
		up(&khelper_sem);
	}
}

#ifdef CONFIG_KMOD
/**
 * call_usermodehelper - start a usermode application
 * @path: pathname for the application
 * @argv: null-terminated argument list
 * @envp: null-terminated environment list
 * @wait: wait for the application to finish and return status.
 *
 * Runs a user-space application.  The application is started
 * asynchronously if wait is not set, and runs as a child of keventd.
 * (ie. it runs with full root capabilities).
 *
 * Must be called from process context.  Returns a negative error code
 * if program was not execed successfully, or 0.
 */
int call_usermodehelper(char *path, char **argv, char **envp, int wait)
{
	DECLARE_COMPLETION(done);
	struct subprocess_info sub_info = {
		.complete	= &done,
		.path		= path,
		.argv		= argv,
		.envp		= envp,
		.wait		= wait,
		.retval		= 0,
	};

	if (!khelper_wq)
		return -EBUSY;

	if (path[0] == '\0')
		return 0;

	if (wait) {
		DECLARE_WORK(work, __call_usermodehelper, &sub_info);
		queue_work(khelper_wq, &work);
	} else {
		DECLARE_WORK(work, __call_usermodehelper_async, &sub_info);
		queue_work(khelper_wq, &work);
	}
	wait_for_completion(&done);
	return sub_info.retval;
}
EXPORT_SYMBOL(call_usermodehelper);

void __init usermodehelper_init(void)
{
	khelper_wq = create_singlethread_workqueue("khelper");
	BUG_ON(!khelper_wq);

	/*
	 * Limit the max number of concurrent processes
	 * to something sane; 10 - max_threads/2 seems ok.
	 */
	if (khelper_max > max_threads/2)
		khelper_max = max_threads/2;
	if (khelper_max < 10)
		khelper_max = 10;

	printk(KERN_INFO "khelper: max %d concurrent processes\n",
	       khelper_max);
	sema_init(&khelper_sem, khelper_max);
}

/*
 * Sanity check for khelper_max is done in usermodehelper_init,
 * as at this point of time the system is not fully initialised.
 */
static int __init khelper_setup(char *s)
{
	get_option(&s, &khelper_max);

	return 1;
}
__setup("khelper_max=", khelper_setup);
#else
int call_usermodehelper(char *path, char **argv, char **envp, int wait)
{
	return -EBUSY;
}
EXPORT_SYMBOL(call_usermodehelper);
#endif
